/***************************************************************************
                          adjacencymatrix.cpp  -  description
                             -------------------
    begin                : Wed Mar 14 2001
    copyright            : (C) 2001 by Gleiss Petra Manuela
    email                : petra@gleiss.at
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "adjacencymatrix.h"
#include "trace.h"

#include <algorithm>

namespace graphlib {

//---------------------------------------------------------------------------
// Con-/Destructors
//---------------------------------------------------------------------------
AdjacencyMatrix::AdjacencyMatrix(const Graph& g) : sizeM(g.numberNodes())
{
  if (sizeM == 0)
    THROW(runtime_error,
        "AdjacencyMatrix::AdjacencyMatrix(const Graph&) empty graph")

  // initialize a map with nodes and unsigned
  makeNodeNumberMap(g);

  // initialize adjacency matrix
  makeAdjacency(g,nodes);

  // initialize degree matrix
  makeDegree(g,nodes);

  //if (nodes) delete nodes;
}

AdjacencyMatrix::AdjacencyMatrix(const AdjacencyMatrix& am)
{
  copy(am);
}

AdjacencyMatrix::~AdjacencyMatrix()
{
  if (adjacencyM) delete[] adjacencyM;
  if (degreeM)    delete[] degreeM;
}

//---------------------------------------------------------------------------
// Operator
//---------------------------------------------------------------------------
AdjacencyMatrix& AdjacencyMatrix::operator=(const AdjacencyMatrix& am)
{
  if (this == &am) return *this;
  copy(am);
  return *this;
}

//---------------------------------------------------------------------------
// In-/Output
//---------------------------------------------------------------------------
ostream& operator<<(ostream& os,const AdjacencyMatrix& l)
{
  os << "\nAdjacency Matrix\n";
  for (unsigned i=0; i!=l.sizeM; i++)
  {
    for (unsigned j=0; j!=l.sizeM; j++)
      os << (int)l.adjacencyM[l.sizeM*i+j] << "   ";
    os << endl;
  }

  os << "\nDegree Matrix\n";
  for (unsigned i=0; i!=l.sizeM; i++)
  {
    for (unsigned j=0; j!=l.sizeM; j++)
      os << (int)l.degreeM[l.sizeM*i+j] << "   ";
    os << endl;
  }
  return os;
}

void AdjacencyMatrix::writeMathematica(ostream& os)
{
  os << "adja = { { ";

  for (unsigned i=0; i!=sizeM; i++)
  {
    for (unsigned j=0; j!=sizeM; j++)
    {
      os << (int)adjacencyM[sizeM*i+j];

      if (j!=sizeM-1)
        os << " , ";
    }
    if (i!=sizeM-1)
      os << " } , { ";
  }
  os << " } }; " << endl << endl;

  os << "degr = DiagonalMatrix [ { ";

  for (unsigned i=0; i!=sizeM; i++)
  {
    os << (int)degreeM[sizeM*i+i];
    if (i!=sizeM-1)
      os << " , ";
  }
  os << " } ]; " << endl << endl;
}

//---------------------------------------------------------------------------
// Utitlities
//---------------------------------------------------------------------------
void AdjacencyMatrix::copy(const AdjacencyMatrix& am)
{
  if (!am.adjacencyM || !am.degreeM)
    THROW(runtime_error,"AdjacencyMatrix::copy(const AdjacencyMatrix&)\
      adjacency or degree == NULL")

  if (adjacencyM) delete[] adjacencyM;
  if (degreeM)    delete[] degreeM;

  sizeM = am.sizeM;
  adjacencyM = new double[sizeM*sizeM];
  degreeM = new double[sizeM*sizeM];


}


void AdjacencyMatrix::makeNodeNumberMap(const Graph& g)
{
  //NodeNumberMap *nodes = new NodeNumberMap();

  constNodeIter it  = g.nodesBegin();
  constNodeIter end = g.nodesEnd();
  unsigned  number = 0;

  // iterate through all nodes of the graph
  while (it != end)
  {
    nodes.insert(nodes.end(),make_pair(*it,number++));
    it++;
  }
  //return nodes;
}

void AdjacencyMatrix::makeAdjacency(const Graph& g,
                                    const NodeNumberMap& nodes)
{
  adjacencyM = new double[sizeM*sizeM];
  fill(adjacencyM,adjacencyM+sizeM*sizeM,0);

  constEdgeIter it  = g.edgesBegin();
  constEdgeIter end = g.edgesEnd();
  unsigned source, target;

  // iterate through all nodes of the graph
  while (it != end)
  {
    source = nodes.find(it->source())->second;
    target = nodes.find(it->target())->second;

    // source -> target
    adjacencyM[source + target*sizeM] = 1;
    // target -> source
    adjacencyM[target + source*sizeM] = 1;
    it++;
  }
}

void AdjacencyMatrix::makeDegree(const Graph& g,
                                 const NodeNumberMap& nodes)
{
  degreeM = new double[sizeM*sizeM];
  fill(degreeM,degreeM+sizeM*sizeM,0);

  constNodeIter  it  = g.nodesBegin();
  constNodeIter end = g.nodesEnd();
  unsigned  number = 0;

  while (it != end)
  {
    number = nodes.find(*it)->second;
    degreeM[number + number*sizeM] = it->degree();
    it++;
  }
}

} // namespace graphlib

/* end of file */
