/***************************************************************************
                          adjacencymatrix.h  -  description
                             -------------------
    begin                : Wed Mar 14 2001
    copyright            : (C) 2001 by Gleiss Petra Manuela
    email                : petra@gleiss.at
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef ADJACENCYMATRIX_H
#define ADJACENCYMATRIX_H

#include "graph.h"
#include <map>

namespace graphlib {

typedef map<Node,unsigned>  NodeNumberMap;

/**
  * @short adjacency matrix and diagoal matrix of vertex degrees.
  */

class AdjacencyMatrix
{
  	//------------------------------------------------------ friend classes
    friend class LaplacianMatrix;

  	//------------------------------------------------------ friend functions
    /**
      * Output-operator.
      *
      * @param os the output-stream
      * @param a AdjacencMatrix
      * @return the output-stream
      */
    friend ostream& operator<<(ostream& os,const AdjacencyMatrix& a);

  public:
  	//------------------------------------------------------ Con-/Destructors
    /**
      * Constructor. Creates new adjacency matrix and diagonal matrix
      * of vertex degrees of <code>graph</code>.
      * <em>Please note:</em> <code>graph</code> may not by empty </em>,
      * otherwise a runtime_error will be thrown.
      *
      * @param graph
      */
    AdjacencyMatrix(const Graph& graph);

    /**
      * CopyConstructor. Makes a copy of the original AdjacencyMatrix.
      * Allocates new memory!
      */
    AdjacencyMatrix(const AdjacencyMatrix&);

    /**
      * Destructor. Deletes both matrices.
      */
    ~AdjacencyMatrix();

  	//------------------------------------------------------ Operators
    /**
      * Assignement operator. Assigns a copy of the original
      * AdjacencyMatrix. Allocates new memory!
      */
    AdjacencyMatrix& operator=(const AdjacencyMatrix&);

  	//------------------------------------------------------ Output
  	/**
  	  * Writes the adjacency and degree matrix to <code>out</code>.
  	  * Format: mathematic
  	  * adja = { { .. , .. } , { .. , .. } };
  	  * degr = DiagonalMatrix [ { .. , .. } ];
  	  *
  	  * @param ostream (default stdout)
  	  */
  	void writeMathematica(ostream&);
  	
  	//------------------------------------------------------ Information
  	/**
  	  * Returns the size of the matrix.
  	  */
    unsigned size() const { return sizeM; }

  	/**
  	  * Returns the size of the matrix.
  	  */
    const double* adjacency() const { return adjacencyM; }

  protected:
  	//------------------------------------------------------ datas
  	/// adjacency matrix
  	double          *adjacencyM;
  	/// diagonal matrix of vertex degrees
  	double          *degreeM;
  	/// number of vertices
  	unsigned        sizeM;
  	/// map node <-> unsigned
  	NodeNumberMap   nodes;

  	//------------------------------------------------------ Utilities
    /**
      * Makes a copy of <code>adjaMat</code>.
      *
      * @param adjaMat
      */
    void copy(const AdjacencyMatrix& adjaMat);

    /**
      * Creates a map so that each node is associated with an
      * ascending number
      *
      * @param graph
      * @return map in which each node is associated with an ascending number
      */
    void makeNodeNumberMap(const Graph&);

    /**
      * Creates the adjacency matrix
      *
      * @param graph
      * @param map in which each node is associated with an ascending number
      */
    void makeAdjacency(const Graph&, const NodeNumberMap&);

    /**
      * Creates the diagonal matrix of the vertices
      *
      * @param graph
      * @param map in which each node is associated with an ascending number
      */
    void makeDegree(const Graph&, const NodeNumberMap&);

}; // class AdjacencyMatrix

} // namespace graphlib

#endif // ADJACENCYMATRIX_H

/* end of file */
