/***************************************************************************
                          bicomponents.cpp  -  description
                             -------------------
    begin                : Thu Mar 8 2001
    copyright            : (C) 2001 by Gleiss Petra Manuela
    email                : petra@gleiss.at
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "bicomponents.h"

#include "nodedata.h"
#include "nodeid.h"

#include "edgedata.h"
#include "edgeid.h"

#include "trace.h"
#include "utilities.h"

#include <algorithm>

namespace graphlib {

//---------------------------------------------------------------------------
// Con-/Destructors
//---------------------------------------------------------------------------
BiComponents::BiComponents(const Graph& g) : bicomp(list<Graph*>()),
                                             comp(list<Graph*>()),
                                             nodeMap(NodeMap()),
                                             edgeMap(EdgeMap())
{
  if (g.numberNodes()==0)
    THROW(runtime_error,
      "BiComponents::BiComponents(const Graph&): empty node set")
  if (g.numberEdges()==0)
    THROW(runtime_error,
      "BiComponents::BiComponents(const Graph&): empty edge set")

    // creates the map <Node,Data for Algorithm>
  makeNodeMap(g);

  // creates the map <Edge,Data for Algorithm>
  makeEdgeMap(g);

  // biconnectivity algorithm
  graph2BiconnComp(g);
}

BiComponents::~BiComponents()
{
  // deleting connected components
  GraphIter it = comp.begin();
  while (it != comp.end())
  {
    if (*it) delete *it++;
  }

  // deleting biconnected components
  it = bicomp.begin();
  while (it!= bicomp.end())
  {
    if (*it) delete *it++;
  }
}

//---------------------------------------------------------------------------
// Output
//---------------------------------------------------------------------------
ostream& operator<<(ostream& os, const BiComponents& bC)
{
  constGraphIter g;

  os << "################### connected components ###################\n";
  for (g=bC.conCompBegin(); g!=bC.conCompEnd(); g++)
    os << **g << endl << endl << endl;

  os << "################### biconnected components ###################\n";
  for (g=bC.biCompBegin(); g!=bC.biCompEnd(); g++)
    os << **g << endl << endl << endl;

  return os;
}

void BiComponents::map_ausgabe(ostream& os) const
{
  os << endl << "NodeMap" << endl;
  NodeMap::const_iterator it = nodeMap.begin();
  for (;it != nodeMap.end(); it++)
    os << it->first.id() << "\t" << it->second << endl;

  os << endl << "NodeSet" << endl;
  NodeSet::const_iterator its = nodeSet.begin();
  for (;its != nodeSet.end(); its++)
    os << its->id() << "\t" << endl;

  os << endl << "EdgeMap" << endl;
  EdgeMap::const_iterator iter = edgeMap.begin();
  for (;iter != edgeMap.end(); iter++)
    os << iter->first << "\t\t"
       << (iter->second ? "visited" : "not visited") << endl;
}

//---------------------------------------------------------------------------
// Utilities
//---------------------------------------------------------------------------
void BiComponents::graph2BiconnComp(const Graph& g)
{
  unsigned step = 1;
  vector<Edge> edge_stack;
  vector<Edge> conn_edges;
  Node v, w, father;
  constEdgeIter e;
  Graph *conn;

  // root of DFS
  v = *(g.nodesBegin());
  nodeMap[v].setBiCompData(true,step,step);
  nodeSet.erase(v);

  for (;;)
  {
//    // single node without any edges
//    if (v.degree() == 0)
//    {
//      // new component
//      conn = new Graph(v); comp.push_back(conn);
//      // new biconnected component
//      Graph *bicon = new Graph(v); bicomp.push_back(bicon);
//      // new current node
//      v = *(nodeSet.begin());
//      nodeMap[v].setBiCompData(true,step,step);
//      nodeSet.erase(v);
//    }
    //
    // forall edges incident to v (current node)
    //
    if ((e=findUnvisitedEdge(v)) != v.adjEnd())
    {
      // push edge e(v,w) into the stack and mark as visited
      edge_stack.push_back(*e);
      conn_edges.push_back(*e);
      edgeMap[*e] = true;
      w = (v.opposite(*e));

      // w already marked
      if (nodeMap[w].isMarked())
        nodeMap[v].changeLow(minimum(nodeMap[v].getLow(),nodeMap[w].getDNF()));
      // w not marked
      else
      {
        step++;
        nodeMap[w].setBiCompData(true,step,step,v);
        v = w;
        nodeSet.erase(v);
      }
    }

    //
    // all edges of v visited
    //
    else
    {
      // father(v)
      father=nodeMap[v].getFather();

      // v = root of DFS
      if (father.isNULL())
      {
        if (!edge_stack.empty())
          THROW(logic_error,
            "BiComponents::Graph2BiconnComp(const Graph&) stack not empty")

        // all nodes marked (last component) -> End
        if (nodeSet.empty())
        {
          conn = new Graph();
          conn->makeSubgraph(conn_edges.begin(),conn_edges.end());
          comp.push_back(conn);
          // comp.push_back(new Graph(g)); // so wrong !!
          conn_edges.clear();
          return;  // END
        }

        // another component
        else
        {
          TRACE("Graph is not connected")

          // connected component - all visited edges
          conn = new Graph();
          conn->makeSubgraph(conn_edges.begin(),conn_edges.end());
          comp.push_back(conn);
          conn_edges.clear();

          v = *(nodeSet.begin());
          nodeMap[v].setBiCompData(true,step,step);
          nodeSet.erase(v);
        }
      }

      // there is a father of v
      else
      {
        // found a component
        if (nodeMap[v].getLow() >= nodeMap[father].getDNF())
        {
          stack2Component(edge_stack, father,v);
        }

        //
        nodeMap[father].changeLow(minimum(nodeMap[v].getLow(),
                                          nodeMap[father].getLow()));

        v = father;
      }
    }
  }
}

void BiComponents::makeNodeMap(const Graph& g)
{
  constNodeIter  it  = g.nodesBegin();
  constNodeIter  end = g.nodesEnd();

  BiCompData data;

  while (it != end)
  {
    nodeMap.insert(nodeMap.end(),make_pair(*it,data));
    nodeSet.insert(nodeSet.end(),*it);
    it++;
  }
}

void BiComponents::makeEdgeMap(const Graph& g)
{
  constEdgeIter it  = g.edgesBegin();
  constEdgeIter end = g.edgesEnd();

  while (it != end)
  {
    edgeMap.insert(edgeMap.end(),make_pair(*it,false));
    it++;
  }
}

//---------------------------------------------------------------------------
// Internal
//---------------------------------------------------------------------------
constEdgeIter BiComponents::findUnvisitedEdge(const Node& n) const
{
  if (n.isNULL())
    THROW(logic_error,
      "BiComponents::findUnvisitedEdge(const Node&) const: node has no data.")

  constEdgeIter it  = n.adjBegin();
  constEdgeIter end = n.adjEnd();

  while (it != end)
  {
    // if edge unvisited -> return
    if (edgeMap.find(*it)->second == false)
      return it;
    else
      it++;
  }
  return it;
}

void BiComponents::stack2Component(vector<Edge>& edge_stack,
                                const Node& father, const Node& v, bool all)
{
  TRACE("biconnected component found");
  Graph *g = new Graph();

  // take the whole stack
  if (all)
  {
    g->makeSubgraph(edge_stack.begin(),edge_stack.end());
    edge_stack.clear();
  }

  // take all edges from the edge(father,v) to the end
  else
  {
    vector<Edge>::iterator begin;

    // search for the edge (father,v) resp. e(v,father)
    begin = find_if(edge_stack.begin(),edge_stack.end(),
                    EdgeSourceTarget(father.id(),v.id()));

    //  make subgraph
    g->makeSubgraph(begin,edge_stack.end());

    // remove "used" edges
    edge_stack.erase(begin,edge_stack.end());
  }

  // add bicomp to the list
  bicomp.push_back(g);
}  	




} // namespace graphlib

/* end of file */
