/***************************************************************************
                          bicomponents.h  -  description
                             -------------------
    begin                : Thu Mar 8 2001
    copyright            : (C) 2001 by Gleiss Petra Manuela
    email                : petra@gleiss.at
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef BICOMPONENTS_H
#define BICOMPONENTS_H

#include "graph.h"
#include "bicompdata.h"

#include <list>
#include <vector>
#include <map>
#include <set>
#include <iostream>

namespace graphlib {

typedef list<Graph*>::iterator               GraphIter;
typedef list<Graph*>::const_iterator         constGraphIter;
typedef map<Node,BiCompData,less<Node> >     NodeMap;
typedef map<Edge,bool,less<Edge> >           EdgeMap;
typedef set<Node>                            NodeSet;

/**
  * @short Connected and Biconnected Components of an undirected graph
  *
  * Biconnected components are maximal beconnected subgraphs of a given graph.
  * @see Graph
  */

class BiComponents
{
  	//------------------------------------------------------ friend functions
    /**
      * Output-operator.
      * Creates an GML output of the components. Each of them is written
      * as a GML graph.
      *
      * @param os the output-stream
      * @param bc the biconnected components
      * @see <a href="http://www.infosun.fmi.uni-passau.de/Graphlet/
      *       GML/index.html">The GML File Format</a>
      * @return the output-stream
      */
    friend ostream& operator<<(ostream&,const BiComponents&);

  public:
  	//------------------------------------------------------ Con-/Destructors
    /**
      * Constructor: Creates all biconnected components <code>g</code>.
      *
      * For each component a new graph is created.
      *
      * @param <code>g</code> the graph
      */
  	BiComponents(const Graph& g);
  	
  	/**
  	  * Destructor: Deletes all components.
  	  */
  	~BiComponents();
   	 	
  	//------------------------------------------------------ Information
  	/**
  	  * Returns the list of biconnected components.
  	  */
  	list<Graph*>& biComp() { return bicomp; }
  	
  	/**
  	  * Returns the number of biconnected components.
  	  */
  	unsigned numberBiComp() const { return bicomp.size(); }
  	
  	/**
  	  * Returns the start of the list of biconnected components.
  	  */
  	constGraphIter biCompBegin() const { return bicomp.begin(); }
  	
  	/**
  	  * Returns the end of the list of biconnected components.
  	  */
  	constGraphIter biCompEnd() const { return bicomp.end(); }
  	
  	/**
  	  * Returns the list of connected components
  	  */
  	list<Graph*>& conComp() { return comp; }
  	
  	/**
  	  * Returns the number of connected components.
  	  */
  	unsigned numberConComp() const { return comp.size(); }
  	  	  	  	
  	/**
  	  * Returns the start of the list of connected components.
  	  */
  	constGraphIter conCompBegin() const { return comp.begin(); }
  	
  	/**
  	  * Returns the end of the list of connected components.
  	  */
  	constGraphIter conCompEnd() const { return comp.end(); }
  	
  	//------------------------------------------------------ Output
    /**
      * writes to ostream the nodeMap and edgeMap.
      *
      * @return ostream
      */
    void map_ausgabe(ostream&) const;

  private:
  	//------------------------------------------------------ Data
    list<Graph*>              bicomp;
    list<Graph*>              comp;
    NodeSet                   nodeSet;          // not marked nodes
    NodeMap                   nodeMap;
    EdgeMap                   edgeMap;

  	//------------------------------------------------------ Utilities
  	/**
  	  * Algorithm: finds all biconnected components and stores them in
  	  * a list of graphs.
  	  */
    void graph2BiconnComp(const Graph&);

  	/**
  	  * Initialize the NodeMap and the NodeSet
  	  */
    void makeNodeMap(const Graph&);

  	/**
  	  * Initialize the EdgeMap
  	  */
    void makeEdgeMap(const Graph&);

    /**
      * Find next unvisited edge incident to node <code>n</code>
      *
      * @return Iterator to found edge
      */
    constEdgeIter findUnvisitedEdge(const Node& n) const;

   /**
      * Creates from all the edges from the top of <code>edge_stack</code>
      * up to the edge (<code>father</code>,<code>v</code>) a new graph.
      * Pushes this graph into components and removes all used edges from
      * the <code>edge_stack</code>. If <code>all</code> is true, all
      * edges will be used.
      * New biconnected component is added to list of components.
      *
      * @param <code>edge_stack</code> stack with edges.
      * @param <code>father</code>
      * @param <code>v</code>
      * @param <code>all</code>
      */
    void stack2Component(vector<Edge>& edge_stack,
                           const Node& father, const Node& v, bool all=false);



}; // class BiComponents

} // namespace graphlib
#endif // BICOMPONENTS_H

/* end of file */
