/***************************************************************************
                          cyclegraph.h  -  description
                             -------------------
    begin                : Mon May 21 2001
    copyright            : (C) 2001 by Gleiss Petra Manuela
    email                : petra@gleiss.at
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef CYCLEGRAPH_H
#define CYCLEGRAPH_H

#include "graph.h"

#include "nodepath.h"


#include <cmath>

#include <vector>
#include <map>
#include <set>

namespace graphlib {

typedef map<Node,NodePath,less<Node> >        NodePathMap2;
typedef set<Node>                             NodeSet;


/**
  *@short An undirected graph
  *
  * A graph for calculating cycle decompositions.
  * Contains functions
  * for the calculation of the minimal cycle basis, the
  * relevant, essential, shortest and unique shortest cyles and the
  * interchangeability classes. These calculations are although
  * possible over the vector space containing the u-paths and cycles,
  * if a list of U-Nodes is set.
  *
  * @see Graph
  * @see
  *   Gleiss, P.M., Leydold, J. and Stadler, P.F. Interchangeability of
  *     Relevant Cycles in Graphs, Elec. J. Comb. 7, R16 (2000)
  *   Vismara, P Union of all the minimum cycle bases of a graph,
  *     Elec. J. Comb. 4, R9 (1997)
  *   <a href="http://www.combinatorics.org/">Elec. J. Comb.</a>
  */

class CycleGraph : virtual public Graph
{

  public:
    //------------------------------------------------------ Con-/Destructors
    /**
      * Constructor.
      */
    CycleGraph();

    /**
      * Constructor.
      */
    CycleGraph(const Graph& g):Graph(g) {;}

 public:
    //------------------------------------------------------ Calculations
    /**
      * Calculates the all shortest pathes from <code>node</code>
      * to all other nodes of the biconnected components represented
      * by <code>set</code> of this graph.
      * The path contains the starting and the end vertices.
      * If <code>ordered</code> is true
      * (\em default), than only pathes to preceding numbers are
      * calculated, else all shortest pathes are calculated.
      * Stores the pathes in a the NodePathMap <code>map</code> and returns
      * the map
      *
      * @param Node
      * @param NodeSet
      * @param NodePathMap
      * @param ordered
      * @return NodePathMap
      */
    NodePathMap2& dijkstra(Node, NodePathMap2& map, NodeSet& set,
                          bool ordered) const;

    /**
      * Initialize <code>map</code>:
      * clears path and pushes all nodes of the Component <code>biComp</code>
      * into it, set distance to INf
      * and marked to false for all nodes except for <code>node</code>,
      * for <code>node</code> distance = 0.
      * Initialize <code>set</code>:
      * pushes all nodes of the list of nodes <code>biComp</code> into it.
      *
      * @param Node
      * @param biComp
      * @param NodePathMap (will be cleared)
      * @param NodeSet (will be cleared)
      */
    void initNodeMap(const Node& node, const Graph& biComp,
                     NodePathMap2& map, NodeSet& set) const;
 private:
    /**
      * Returns a node from the <code>nodeSet</code>
      * with the smallest distance found in <code>nodeMap</code>.
      *
      * @param const NodePathMap
      * @param const NodeSet
      * @return Node
      */
    Node findNodeWithSmallestDistance(const NodePathMap2& nodeMap,
         const NodeSet&) const;



  protected:
    //------------------------------------------------------ Datas


  private:
    /// for debugging -> writes: [node.id] distance marked "path:" path
    void writeNodeMap(const NodePathMap2&, ostream&) const;



}; // class CycleGraph

} // namespace graphlib

#endif // CYCLEGRAPH_H

/* end of file */
