/***************************************************************************
                          eigenpair.cpp  -  description
                             -------------------
    begin                : Wed Mar 14 2001
    copyright            : (C) 2001 by Gleiss Petra Manuela
    email                : petra@gleiss.at
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "eigenpair.h"
#include "node.h"

#include "trace.h"

#include <algorithm>
#include <iomanip>

namespace graphlib {

//---------------------------------------------------------------------------
// Con-/Destructors
//---------------------------------------------------------------------------
EigenPair::EigenPair(unsigned size) : eigenvalue(0.0), dim(size),
                                      weakND(-1), strongND(-1),
                                      nEvMap(0)
{
  eigenvector = new double[dim];
  fill(eigenvector,eigenvector+dim,0);
  maximum = minimum = 0;
}

EigenPair::EigenPair(double value, double* matrix, unsigned dimension, unsigned col) :
      eigenvalue(value), dim(dimension), weakND(-1), strongND(-1),
      nEvMap(0)
{
  if (col > dim)
    THROW(range_error,"EigenPair::EigenPair(double, double*, unsigned, unsigned)\
      column out of range")

  eigenvector = new double[dim];
  maximum = minimum = 0;

  // vector == col-th column of the matrix
  for (unsigned i=0; i!=dim; i++)
  {
    eigenvector[i] = matrix[i*dim+col];
    if (eigenvector[i] > maximum)
      maximum = eigenvector[i];
    else if (eigenvector[i] < minimum)
      minimum = eigenvector[i];
  }
}

EigenPair::EigenPair(const list<EigenPair*>& eigenPL,
                     const vector<double>& factors)
                : weakND(-1), strongND(-1), nEvMap(0)
{
  constEigenPairIter it = eigenPL.begin();
  constEigenPairIter end = eigenPL.end();

  vector<double>::const_iterator fit = factors.begin();
  vector<double>::const_iterator fend = factors.end();

  eigenvalue = (*it)->eigenvalue;
  dim  = (*it)->dim;

  eigenvector = new double[dim];
  fill(eigenvector,eigenvector+dim,0);
  maximum = minimum = 0;

  for (unsigned i=0; i!=dim; i++)
  {
    for (it=eigenPL.begin(),fit=factors.begin();it!=end, fit!=fend; it++, fit++)
      {
        eigenvector[i] += (*it)->eigenvector[i] * (*fit);
      }

    if (eigenvector[i] > maximum)
      maximum = eigenvector[i];
    else if (eigenvector[i] < minimum)
      minimum = eigenvector[i];
  }
}


EigenPair::~EigenPair()
{
  if (eigenvector) delete[] eigenvector;
  if (nEvMap) delete nEvMap;
}

//---------------------------------------------------------------------------
// Operators
//---------------------------------------------------------------------------
EigenPair& EigenPair::operator= (const EigenPair& ep)
{
  if (this == &ep)
    return *this;

  eigenvalue = ep.eigenvalue;
  dim  = ep.dim;

  if (eigenvector) delete[] eigenvector;
  eigenvector = new double[dim];
  for (unsigned i=0; i!=dim; i++)
    eigenvector[i] = ep.eigenvector[i];
  minimum = ep.minimum;
  maximum = ep.maximum;

  return *this;
}

bool EigenPair::operator== (const EigenPair& ep) const
{
  if (this == &ep)
    return true;

  if (dim != ep.dim)
    THROW(runtime_error,
        "EigenPair::operator==(const EigenPair&) different dimension.")

  if (eigenvalue != ep.eigenvalue)
    return false;

  for (unsigned i=0; i!=dim; i++)
    if (eigenvector[i] != ep.eigenvector[i])
      return false;

  return true;
}

bool EigenPair::operator< (const EigenPair& ep) const
{
  if (dim != ep.dim)
    THROW(runtime_error,
        "EigenPair::operator==(const EigenPair&) different dimensions.")

  return (eigenvalue < ep.eigenvalue ? true : false);
}

//---------------------------------------------------------------------------
// In-/Output
//---------------------------------------------------------------------------
ostream& operator<<(ostream& os,const EigenPair& eP)
{
  os.precision(16);
  os.setf(ios::showpos | ios::fixed);

  os << setw(20) << eP.eigenValue() << " :  ";
  for (unsigned i=0; i!=eP.dimension(); i++)
    os << setw(20) << (eP.eigenVector())[i] << "   ";

  return os;
}

//---------------------------------------------------------------------------
// Calculations
//---------------------------------------------------------------------------
int EigenPair::calcWeakNodalDomain(const list<Node>& nodes)
{
  // initialize the map : nodes <-> value of eigenvector / number of nodal domain
  map<Node,int>* nodalDomain = makeNodeEigenVectorMap(nodes);
  weakND = 0;



  if (nodalDomain) delete nodalDomain;

  return weakND;
}

int EigenPair::calcStrongNodalDomain(const list<Node>& nodes)
{
  // initialize the map : nodes <-> value of eigenvector / number of nodal domain
  map<Node,int>* nodalDomain =   makeNodeEigenVectorMap(nodes);
  strongND = 0;




  if (nodalDomain) delete nodalDomain;

  return strongND;
}

map<Node,int>* EigenPair::makeNodeEigenVectorMap(const list<Node>& nodes)
{
  // delete old
  if (nEvMap) delete nEvMap;

  // map for node <-> nodalDomain
  map<Node, int>* nodalDomain = new map<Node,int>;

  // map for node <-> value of Eigenvector
  nEvMap = new map<Node,double>;

  constNodeIter it = nodes.begin();
  constNodeIter end= nodes.end();
  unsigned i=0;

  // initializing map
  while (it != end)
  {
    nodalDomain->insert(nodalDomain->end(),make_pair(*it,0));
    nEvMap->insert(nEvMap->end(),make_pair(*it++,eigenvector[i++]));
  }

  return nodalDomain;
}

} // namespace graphlib

/* end of file */
