/***************************************************************************
                          eigenvector_eigenvalue.cpp  -  description
                             -------------------
    begin                : Wed Mar 14 2001
    copyright            : (C) 2001 by Gleiss Petra Manuela
    email                : petra@gleiss.at
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "eigenvector_eigenvalue.h"
#include "trace.h"
#include "utilities.h"

#include "eigen.h"



namespace graphlib {

//---------------------------------------------------------------------------
// Con-/Destructors
//---------------------------------------------------------------------------
Eigenvector_Eigenvalue::Eigenvector_Eigenvalue(const Graph& g):sort(true)
{
  string error = "Eigenvector_Eigenvalue::Eigenvector_Eigenvalue(const Graph&)";
  if (g.numberNodes() == 0)
    THROW(runtime_error,error+" empty graph")

  if (g.numberNodes() == 1)
    THROW(runtime_error,error+" graph must have more than one node")

  // make laplacian matrix
  LaplacianMatrix l(g);

  unsigned size = l.sizeM;

  // copy laplacian matrix
  double *eigenvector = new double[size*size];
  for (unsigned i=0; i!=size*size; i++)
    eigenvector[i] = l.laplacianM[i];

  // calculates eigenvalues and eigenvectors
  makeEigen(eigenvector, size);
}

Eigenvector_Eigenvalue::Eigenvector_Eigenvalue(const LaplacianMatrix* l, double* orth):sort(true)
{
  if (! l)
    THROW(runtime_error,"Eigenvector_Eigenvalue::Eigenvector_Eigenvalue\
        (const LaplacianMatrix*) LaplacianMatrix==NULL")

  unsigned size = l->sizeM;

  // copy laplacian matrix
  double *eigenvector = new double[size*size];
  for (unsigned i=0; i!=size*size; i++)
    eigenvector[i] = l->laplacianM[i];

  // calculates eigenvalues and eigenvectors
  makeEigen(eigenvector, size, orth);
}

Eigenvector_Eigenvalue::Eigenvector_Eigenvalue(double* eigenvector, 
					       unsigned size):sort(false)
{
  // calculates eigenvalues and eigenvectors
  makeEigen(eigenvector, size);
}

Eigenvector_Eigenvalue::~Eigenvector_Eigenvalue()
{
  list<EigenPair*>::iterator i;
  if (!eigenpairs.empty())
    for (i=eigenpairs.begin();i!=eigenpairs.end();i++)
      delete *i;
}

//---------------------------------------------------------------------------
// In-/Output
//---------------------------------------------------------------------------
ostream& operator<<(ostream& os,const Eigenvector_Eigenvalue& eV)
{
  constEigenPairIter it=eV.eigenpairs.begin();
  unsigned dim = (*it)->dimension();

  os.precision(4);
  os.setf(ios::showpos | ios::fixed);

  for (it=eV.eigenpairs.begin();it!=eV.eigenpairs.end();it++)
    os << setw(10) << (*it)->eigenValue() << "\t";
  os << endl << endl;

  for (unsigned i=0; i!=dim; i++)
  {
    for (it=eV.eigenpairs.begin();it!=eV.eigenpairs.end();it++)
      os << setw(10) << ((*it)->eigenVector())[i] << "\t";
    os << endl;
  }


  return os;
}

void Eigenvector_Eigenvalue::writeMathematica(ostream& os)
{
  constEigenPairIter it=eigenpairs.begin();
  constEigenPairIter end=eigenpairs.end();
  unsigned size = (*it)->dimension();
  unsigned vorletztes = size-1;
  unsigned j = 0;

  os << "eval = { ";

  for (;it!=end;it++,j++)
  {
    os << toMathematica((*it)->eigenValue());
    if (j != vorletztes)
      os << " , ";
  }
  os << " };"  << endl << endl;

  os << "evec = { { ";
  for (it=eigenpairs.begin(), j=0;it!=end;it++,j++)
  {
    for (unsigned i=0; i!=size; i++)
    {
      os << toMathematica(((*it)->eigenVector())[i]);
      if (i != vorletztes)
        os << " , ";
    }
    if (j != vorletztes)
      os << " } , { ";
  }
  os << " } }; " << endl << endl;
}

//---------------------------------------------------------------------------
// Information
//---------------------------------------------------------------------------
double Eigenvector_Eigenvalue::minEigenValue() const
{
  return (*eigenpairs.begin())->eigenValue();
}
  	
double Eigenvector_Eigenvalue::maxEigenValue() const
{
  constEigenPairIter it = eigenpairs.end();
  it--;
  return (*it)->eigenValue();
}


//---------------------------------------------------------------------------
// Utilities
//---------------------------------------------------------------------------
void Eigenvector_Eigenvalue::makeEigen(double* eigenvector, unsigned size, double* orth)
{
  /*
  unsigned size = l.sizeM;

  // copy laplacian matrix
  double *eigenvector = new double[size*size];
  for (unsigned i=0; i!=size*size; i++)
    eigenvector[i] = l.laplacianM[i];
  */

  double *eigenvalue = new double[size];

// from ccmath -> calculates eigenvector and eigenvalue
  eigen(eigenvector,eigenvalue,size);

  if (orth) 
  {
    double *res = new double[size*size];
    mmul(res, orth, eigenvector, size);
    if (eigenvector) 
      delete [] eigenvector;
    eigenvector = res;
    mnorm(eigenvector,size,size);
  }

  minimal = maximal = 0;

  for (unsigned i=0; i!=size; i++)
  {
    // new EigenPair with eigenvalue and eigenvector (i==column)
    EigenPair *p = new EigenPair(eigenvalue[i],eigenvector,size,i);
    eigenpairs.push_back(p);
    if (minimal > p->maximumEV())
      minimal = p->maximumEV();
    else if (maximal < p->minimumEV())
      maximal = p->minimumEV();
  }
  
  if (sort)
    eigenpairs.sort(lessEigenPairPointer());

  //if (eigenvalue) delete[] eigenvalue;
  //if (eigenvector) delete[] eigenvector;
}

} // namespace graphlib

/* end of file */
