/***************************************************************************
                          eigenvector_eigenvalue.h  -  description
                             -------------------
    begin                : Wed Mar 14 2001
    copyright            : (C) 2001 by Gleiss Petra Manuela
    email                : petra@gleiss.at
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef EIGENVECTOR_EIGENVALUE_H
#define EIGENVECTOR_EIGENVALUE_H

#include "graph.h"
#include "eigenpair.h"
#include "laplacianmatrix.h"

namespace graphlib {


/**
  * @short eigenvalues and eigenvectors of a graph
  *
  * Graphs with only one node, are not allowed. (Bug in ccmath,
  * in function housev()).
  *
  */

class Eigenvector_Eigenvalue
{
  	//------------------------------------------------------ friend functions
    /**
      * Output-operator.
      *
      * @param os the output-stream
      * @param eV list of EigenPair
      * @return the output-stream
      */
    friend ostream& operator<<(ostream& os,const Eigenvector_Eigenvalue& eV);

  public:
  	//------------------------------------------------------ Con-/Destructors
    /**
      * Constructor: Creates a list of EigenPair, holding eigenvectors
      * and eigenvalue of the <code>g</code>.
      * <em>Please note:</em> <code>graph</code> may not by empty </em>,
      * otherwise a runtime_error will be thrown.
      *
      * @param <code>g</code> the graph
      */
  	Eigenvector_Eigenvalue(const Graph& g);
  	
    /**
      * Constructor: Creates a list of EigenPair, holding eigenvectors
      * and eigenvalue of a graph with <code>l</code> as its Laplacian Matrix.
      *
      * @param <code>l</code> the Laplacian of the graph
      */
  	Eigenvector_Eigenvalue(const LaplacianMatrix* l, double* orth = 0);
  	
    /**
      * Constructor: Creates a list of EigenPair, holding eigenvectors
      * and eigenvalue of the matrix <code>eigenvector</code>.
      *
      * @param <code>l</code> the Laplacian of the graph
      */
  	Eigenvector_Eigenvalue(double* eigenvector, unsigned size);

  	/**
  	  * Destructor: Deletes all components.
  	  */
  	~Eigenvector_Eigenvalue();
   	 	
  	//------------------------------------------------------ Output
  	/**
  	  * Writes the adjacency and degree matrix to <code>out</code>.
  	  * Format: mathematic
  	  * eval = { .. , .. };
  	  * evev = { { .. , .. } , { .. , .. } };
  	  *
  	  * eval and evec: format of doubles: 1.xxxxxxxxxxxxxxxx*^exponent
  	  *
  	  * @param ostream (default stdout)
  	  */
  	void writeMathematica(ostream&);
  	
  	//------------------------------------------------------ Information
  	/**
  	  * Returns the list of EigenPairs.
  	  */
  	list<EigenPair*> eigenPairs() { return eigenpairs; }
  	
 	  /**
  	  * Returns the begin of list of EigenPairs.
  	  */
  	constEigenPairIter eigenPairsBegin() const { return eigenpairs.begin(); }
  	
 	  /**
  	  * Returns the end of list of EigenPairs.
  	  */
  	constEigenPairIter eigenPairsEnd() const { return eigenpairs.end(); }
  	
  	/**
  	  * Returns smallest EigenValue.
  	  */
  	double minEigenValue() const;
  	
  	/**
  	  * Returns greatest EigenValue.
  	  */
  	double maxEigenValue() const;
  	
  	/**
  	  * Returns lowest value of EigenVectors.
  	  */
  	double minEigenVector() const { return minimal; }
  	
  	/**
  	  * Returns highest value of EigenVectors.
  	  */
  	double maxEigenVector() const { return maximal; }
  	  	  	  	
  protected:
  	//------------------------------------------------------ Data
    /// list of eigenpairs: eigenvalue + eigenvector
    list<EigenPair*>           eigenpairs;

    /// minimal value of eigenvectors
    double minimal;
    /// maximal value of eigenvectors
    double maximal;

    bool sort;

  	//------------------------------------------------------ Utilities
  	/**
  	  * Calculates Eigenvalues and Eigenvectors
  	  *
  	  * @param Laplacian matrix
  	  */
    void makeEigen(double* eigenvector, unsigned size, double* orth = 0); 	
  	
  	
}; // class Eigenvector_Eigenvalue

} // namespace graphlib
#endif  //EIGENVECTOR_EIGENVALUE_H

/* end of file */
