/***************************************************************************
                          gml_graph.cpp  -  description
                             -------------------
    begin                : Wed Mar 7 2001
    copyright            : (C) 2001 by Gleiss Petra Manuela
    email                : petra@gleiss.at
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "gml_graph.h"

#include "nodedata.h"
#include "nodeid.h"

#include "edgedata.h"
#include "edgeid.h"

#include "trace.h"
#include "fileerror.h"
#include "gml_error.h"

namespace graphlib {

//---------------------------------------------------------------------------
// Con-/Destructors
//---------------------------------------------------------------------------

GML_Graph::GML_Graph(string fname) : Graph(),
                                     filename(fname),
                                     maxCoords(),
                                     minCoords()
{
  in.open(filename.c_str(), ios::in);
  if (in.bad() || in.fail())
    THROW(FileError,"can not open"+filename)

  GML2Graph();
}

//---------------------------------------------------------------------------
// Information
//---------------------------------------------------------------------------

Graph* GML_Graph::graph()
{
  return this;
}
//---------------------------------------------------------------------------
// Utilities
//---------------------------------------------------------------------------
void GML_Graph::GML2Graph()
{
  // find "graph"
  findGraph();

  int count_bracket=0;
  string tmp_string;

  do
  {
    in >> tmp_string;

    // a new node
    if (tmp_string=="node")
      readNode();

    // a new edge
    else if (tmp_string == "edge")
      readEdge();

    // opening brackets
    else if (tmp_string == "[")
      count_bracket++;

    // closing brackets
    else if (tmp_string == "]")
      count_bracket--;

    // comments
    else if (tmp_string == "#")
      getline(in,tmp_string);

    // name of graph
    else if (tmp_string == "name")
      name = tmp_string;

    else
      TRACE(filename+": ignored characters: "+tmp_string)

  } while (!in.eof() && count_bracket!=0);

  if (count_bracket < 0)
    THROW(GML_Error, "to many closing brackets")
  else if (count_bracket > 0)
    THROW(GML_Error, "to many opening brackets")
}

void GML_Graph::findGraph()
{
  string tmp;
  // read till "graph" found or EOF
  do
  {
    in >> tmp;

    // # -> comment
    if (tmp[0] == '#')
      getline(in,tmp);

    // "graph" has to be the first key
    else if (tmp != "graph")
      THROW(GML_Error,"unexpected character before \"graph\"")



  } while (!in.eof() && tmp!="graph" );

  if (in.eof())
    THROW(GML_Error," END OF FILE - no graph found");
}

void GML_Graph::readNode()
{
  int count_bracket=0;
  unsigned id=0;
  string buffer;
  ostringstream err;

  // create new node
  Node n;
  n.data = new NodeData;

  do
  {
    buffer="\0";
    in >> buffer;

    // id of the node
    if (buffer == "id")
    {
      in >> id;
      if  (!in.good())
        THROW(GML_Error,"node id: unexpected character")

      if (getFreeNodeNumber(id) != true)
      {
        err << "node " << id << " already exists";// << ends;
        THROW(GML_Error,err.str())
        //err.freeze(0);
      }
      // node id OK
      n.data->id = id;
    }

    // label or parameter of the node
    else if (buffer == "label" || buffer == "parameter")
    {
      char tmp;
      in >> tmp;
      if (tmp !='"')
      {
        err << "empty label in node " << n.id();// << ends;
        THROW(GML_Error,err.str())
        //err.freeze(0);
      }

      else
        getline(in,n.data->label,'"');
    }

    // graphics of the node
    else if (buffer == "graphics")
      readCoordinates(&n);

    else if (buffer == "[")
      count_bracket++;

    else if (buffer == "]")
      count_bracket--;

    else if (buffer == "node" || buffer == "edge")
    {
      err << " syntax error near node " << n.id();// << ends;
      THROW(GML_Error,err.str())
      //err.freeze(0);
    }

    else
      TRACE(filename+" ignored characters: "+buffer)

  } while (!in.eof() && count_bracket!=0);

  if (count_bracket)
    THROW(GML_Error," unexpected EOF")

  // add to the graph
  n.data->pos=nodes.insert(nodes.end(),n);
  ++nodes_count;

  ostringstream x;
  x << "node: " <<n;// << ends;
  TRACE(x.str());
  //x.freeze(0);
}

void GML_Graph::readCoordinates(Node* n)
{
  int count_bracket=0;
  double tmp;
  string buffer,zeile;
  ostringstream err;
  err << "near node " << n->id();
  do
  {
    buffer="\0";
    in >> buffer;
    if (buffer == "[")
      count_bracket++;

    else if (buffer == "]")
      count_bracket--;

    else if (buffer == "x")
    {
      in >> tmp;
      n->data->coord.x=tmp;
      if (maxCoords.x < tmp) maxCoords.x=tmp;
      else if (minCoords.x > tmp) minCoords.x=tmp;
    }

    else if (buffer == "y")
    {
      in >> tmp;
      n->data->coord.y = tmp;
      if (maxCoords.y < tmp) maxCoords.y=tmp;
      else if (minCoords.y > tmp) minCoords.y=tmp;
    }

    else if (buffer == "z")
    {
      in >> tmp;
      n->data->coord.z = tmp;
      if (maxCoords.z < tmp) maxCoords.z=tmp;
      else if (minCoords.z > tmp) minCoords.z=tmp;
    }

    else if (buffer[0] == '#')
      getline(in,zeile);

    else if (buffer == "node" || buffer == "edge")
    {
      err << " syntax error";// << ends;
      THROW(GML_Error,err.str())
      //err.freeze(0);
    }

    else
      TRACE(filename + " ignored characters: " + buffer)

    if (!in.good())
    {
      err << " graphics: unexpected character";// << ends;
      THROW(GML_Error,err.str())
      //err.freeze(0);
    }


  } while (!in.eof() && count_bracket!=0);

  if (count_bracket)
  {
    err << "bracket error";// << ends;
    THROW(GML_Error,err.str())
    //err.freeze(0);
  }
}

void GML_Graph:: readEdge()
{
  unsigned source=0, target=0;
  string label="";

  int count_bracket=0;
  string buffer;
  ostringstream err;

  do
  {
    buffer="\0";
    in >> buffer;

    if (buffer == "label" || buffer == "parameter")
    {
      char tmp;
      in >> tmp;
      if (tmp !='"')
        THROW(GML_Error,"empty label in edge ")
      else
        getline(in,label,'"');
    }

    else if (buffer == "source")
    {
      in >> source;
      if (!in.good())
        THROW(GML_Error,"edge source: unexpected character")
    }

    else if (buffer == "target")
    {
      in >> target;
      if (!in.good())
        THROW(GML_Error,"edge target: unexpected character")
    }

    else if (buffer == "[")
      count_bracket++;

    else if (buffer == "]")
      count_bracket--;

    else if (buffer == "edge" || buffer == "node")
    {
      err << "syntax error near edge (" << source << ","
          << target << ")";// << ends;
      THROW(GML_Error,err.str())
      //err.freeze(0);
    }

    else
      TRACE(filename+" ignored characters: "+buffer)

  } while (!in.eof() && count_bracket!=0);

  if (count_bracket)
    THROW(GML_Error,"unexpected EOF")

  // create edge
  newEdge(source,target,label);
}

} // namespace graphlib

/* end of file */
