/***************************************************************************
                          gml_parser.c  -  description
                             -------------------
    begin                :
    copyright            : (C) 1997 by Michael Himsolt
    email                : himsolt@fmi.uni-passau.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


/*
 *  parser for the GML-file-format specified in:
 *  Michael Himsolt, GML: Graph Modelling Language, 21.01.1997  
 */ 

#ifndef _GML_PARSER_CPP_
#define _GLM_PARSER_CPP_

#include <cstdio>
#include <cstdlib>
#include <cassert>
#include <cstring>
#include "gml_parser.h"
//#include "utilities.h"
//#include "gml_scanner.cpp" //changed by gil 10/05/02

namespace graphlib {


/**
  * returns list of KEY - VALUE pairs. Errors and a pointer to a list
  * of key-names are returned in GML_stat. Previous information contained
  * in GML_stat, i.e. the key_list, will be *lost*.
  */
template <class X> struct GML_pair* GML_parser (X source, struct GML_stat* stat, int open)
{
  struct GML_token token;
  struct GML_pair* pair;
  struct GML_pair* list;
  struct GML_pair* tmp = NULL;
  struct GML_list_elem* tmp_elem;

  assert (stat);

  pair = (struct GML_pair*) malloc (sizeof (struct GML_pair));
  list = pair;

  for (;;)
  {
	  token = GML_scanner (source); // for GML_pair kind
	
	  // GML_END
	  if (token.kind == GML_END)
	  {
	    if (open)
	    {
	      stat->err.err_num = GML_OPEN_BRACKET;
    		stat->err.line = GML_line;
		    stat->err.column = GML_column;
		    free (pair);

    		if (tmp == NULL)
    		  return NULL;
    		else
    		{
		      tmp->next = NULL;
		      return list;
		    }
	    }
      break;
    }

    // GML_R_BRACKET
    else if (token.kind == GML_R_BRACKET)
    {
	    if (!open)
	    {
	      stat->err.err_num = GML_TOO_MANY_BRACKETS;
		    stat->err.line = GML_line;
    		stat->err.column = GML_column;
    		free (pair);

    		if (tmp == NULL)
    		  return NULL;
		    else
		    {
    		  tmp->next = NULL;
		      return list;
		    }
	    }
	    break;
    }

    // GML_ERROR
    else if (token.kind == GML_ERROR)
    {
      stat->err.err_num = token.value.err.err_num;
	    stat->err.line = token.value.err.line;
	    stat->err.column = token.value.err.column;
	    free (pair);
	      
	    if (tmp == NULL)
		    return NULL;
	    else
	    {
		    tmp->next = NULL;
		    return list;
	    }
    }

    // NOT GML_KEY -> wrong syntax in the file
    else if (token.kind != GML_KEY)
    {
	    stat->err.err_num = GML_SYNTAX;
	    stat->err.line = GML_line;
	    stat->err.column = GML_column;
	    free (pair);
	   
	    if (token.kind == GML_STRING)
		    free (token.value.str);

	    if (tmp == NULL)
		    return NULL;
	    else
	    {
		    tmp->next = NULL;
		    return list;
	    }
	  }
	   
	  // GML_KEY, GML_INT, GML_DOUBLE, GML_STRING, GML_L_BRACKET
	  // first time -> memory allocation, new key_list
	  if (!stat->key_list)
	  {
	    stat->key_list = (struct GML_list_elem*)
	          malloc (sizeof (struct GML_list_elem));
	    stat->key_list->next = NULL;
	    stat->key_list->key = token.value.str;
	    pair->key = token.value.str;
	    pair->lineNr = GML_line;
	  }
	
	  // key_list contains already some elements
	  else
	  {
	    tmp_elem = stat->key_list;
	    tmp_elem->lineNr = GML_line;
	    
	    while (tmp_elem)
	    {
		    if (!strcmp (tmp_elem->key, token.value.str))
		    {
		      free (token.value.str);
		      pair->key = tmp_elem->key;
		      pair->lineNr = tmp_elem->lineNr;
		      break;
		    }
		    tmp_elem = tmp_elem->next;
	    }
	
	    if (!tmp_elem)
	    {
		    tmp_elem = (struct GML_list_elem*)
		    malloc (sizeof (struct GML_list_elem));
    		tmp_elem->next = stat->key_list;
    		stat->key_list = tmp_elem;
    		tmp_elem->key = token.value.str;
    		pair->key = token.value.str;
  	    pair->lineNr = GML_line;
	    }
  	}
	
    token = GML_scanner (source); // for GML_pair value
    pair->lineNr = GML_line;
	
  	switch (token.kind)
  	{
   	case GML_INT:
 	    pair->kind = GML_INT;
 	    pair->value.integer = token.value.integer;
 	    break;

   	case GML_DOUBLE:
 	    pair->kind = GML_DOUBLE;
 	    pair->value.floating = token.value.floating;
 	    break;

   	case GML_STRING:
 	    pair->kind = GML_STRING;
 	    pair->value.str = token.value.str;
 	    break;

   	case GML_L_BRACKET:
 	    pair->kind = GML_LIST;
 	    pair->value.list = GML_parser (source, stat, 1);
     	
 	    if (stat->err.err_num != GML_OK)
 	      return list;

 	    break;

   	case GML_ERROR:
 	    stat->err.err_num = token.value.err.err_num;
 	    stat->err.line = token.value.err.line;
 	    stat->err.column = token.value.err.column;
 	    free (pair);
     		
 	    if (tmp == NULL)
 	      return NULL;
 	    else
 	    {
 		    tmp->next = NULL;
 		    return list;
 	    }
 	    break;

   	default:
 	    stat->err.line = GML_line;
 	    stat->err.column = GML_column;
 	    stat->err.err_num = GML_SYNTAX;
 	    free (pair);

 	    if (tmp == NULL)
 	      return NULL;
 	    else {
     		tmp->next = NULL;
     		return list;
 	    }
    }

  	tmp = pair;
  	pair = (struct GML_pair*) malloc (sizeof (struct GML_pair));
  	tmp->next = pair;
  }

  stat->err.err_num = GML_OK;
  free (pair);

  if (tmp == NULL)
  	return NULL;
  else
  {
  	tmp->next = NULL;
  	return list;
  }
}
	
template struct GML_pair* GML_parser<string*>(string*, struct GML_stat*, int);
template struct GML_pair* GML_parser<char**>(char**, struct GML_stat*, int);

void GML_free_list (struct GML_pair* list, struct GML_list_elem* keys)
{
  struct GML_pair* tmp = list;
  struct GML_list_elem* tmp_key;

  while (keys)
  {
    free (keys->key);
  	tmp_key = keys->next;
	  free (keys);
	  keys = tmp_key;
  }

  while (list)
  {
	  switch (list->kind)
	  {
	  case GML_LIST:
	    GML_free_list (list->value.list, NULL);
	    break;

  	case GML_STRING:
	    free (list->value.str);
	    break;

	  default:
	    break;
	  }
	
	  tmp = list->next;
	  free (list);
	  list = tmp;
  }
}
	
void GML_print_list (struct GML_pair* list, int level)
{
  if (level == 0)
    printf("************************************************\n");
  struct GML_pair* tmp = list;
  int i;

  while (tmp)
  {
    for (i = 0; i < level; i++)
      printf ("    ");

    printf ("*KEY* : %s", tmp->key);

    switch (tmp->kind)
    {
    case GML_INT:
      printf ("  *VALUE* (long) : %ld \n", tmp->value.integer);
    	break;

    case GML_DOUBLE:
      printf ("  *VALUE* (double) : %f \n", tmp->value.floating);
      break;

    case GML_STRING:
      printf ("  *VALUE* (string) : %s \n", tmp->value.str);
      break;
   	
    case GML_LIST:
      printf ("  *VALUE* (list) : \n");
      GML_print_list (tmp->value.list, level+1);
      break;

    default:
      break;
    }
   	
    tmp = tmp->next;
  }
}
	
void GML_printList(struct GML_pair* list, int level, ostream& os)
{
  static string message;

  if (level == 0)
    os << "************* List of GML Key-Value Pairs *************\n";
  struct GML_pair* tmp = list;
  int i;

  while (tmp)
  {
    for (i = 0; i < level; i++)
      os << "    ";

    os << "*KEY* : " << tmp->key;

    switch (tmp->kind)
    {
    case GML_INT:
      os << "  *VALUE* (long) : " << tmp->value.integer << endl;
    	break;

    case GML_DOUBLE:
      os << "  *VALUE* (double) : " << tmp->value.floating << endl;
      break;

    case GML_STRING:
      os << "  *VALUE* (string) : " << tmp->value.str << endl;
      break;
   	
    case GML_LIST:
      os << "  *VALUE* (list) : \n";
      GML_printList (tmp->value.list, level+1, os);
      break;

    default:
      break;
    }
   	
    tmp = tmp->next;
  }
  if (level == 0)
    os << "*******************************************************\n";
}
	
char* GML_print_error (struct GML_stat* stat, const char* filename)
{
  char buffer[500];
  char* error;

  sprintf(buffer,"in line %d column %d ",
      stat->err.line,stat->err.column);

  switch (stat->err.err_num)
  {
  case GML_OK: // valid GML file -> make Graph
    strcat(buffer,"ok!");
    break;

  case GML_UNEXPECTED:
    strcat(buffer,"unexpected charactar");
	  break;
	
  case GML_SYNTAX:
  	strcat(buffer,"syntax error");
  	break;
	
  case GML_PREMATURE_EOF:
  	strcat(buffer,"unexpected EOF");
  	break;
	
  case GML_TOO_MANY_DIGITS:
  	strcat(buffer,"number with too many digits");
  	break;
  	
  case GML_OPEN_BRACKET:
  	strcat(buffer,"open bracket left");
  	break;
  	
  case GML_TOO_MANY_BRACKETS:
  	strcat(buffer,"too many closing brackets");
  	break;
  	
  case GML_EDGE_NODE_UNKNOWN:
  	strcat(buffer,"edge - unknown node");
  	break;
  	
  case GML_EDGE_TWICE:
  	strcat(buffer,"edge twice");
  	break;
  	
  default:
    strcat(buffer,"unknown error");
    break;
  }
  strcat(buffer," in ");
  strcat(buffer,filename);

  error = (char*) malloc(sizeof(buffer));
  strcpy (error,buffer);
  return error;
}

} // namespace graphlib

#endif
/* end of file */






