/***************************************************************************
                          gml_scanner.h  -  description
                             -------------------
    begin                :
    copyright            : (C) 1997 by Michael Himsolt
    email                : himsolt@fmi.uni-passau.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef GML_SCANNER_H
#define GML_SCANNER_H

#include <cstdio>
#include <string>

#include <cassert>
#include <cstring>
#include <cctype>
#include <sstream>
#include <typeinfo>

#include "gml_utils.h"

/*
namespace { //moved from .cpp to .h by gil 090702 for GML_scanner() start
  //  #include "gml_utils.h"

  using namespace graphlib::graphlib_utils;
}
*/

namespace graphlib {

  int GML_search_ISO (char*, int); //added by gil 090702
/*
 * start-size of buffers for reading strings. If too small it will be enlarged
 * dynamically
 */

#define INITIAL_SIZE 1024

/**
  * Possible value types while parsing a GML file
  */
typedef enum
{
    GML_KEY, GML_INT, GML_DOUBLE, GML_STRING, GML_L_BRACKET, 
    GML_R_BRACKET, GML_END, GML_LIST, GML_ERROR
} GML_value; 

/**
  * Possible errors while parsing a GML file.
  * <ul>
  *   <li>
  *     <code>GML_EDGE_NODE_UNKNOWN</code>: source or target of the edge is
  *     unknown.
  *   </li>
  *   <li>
  *     <code>GML_EDGE_TWICE</code>: source or target of the edge is
  *     unknown.
  *   </li>
  *   <li>
  *     <code>GML_FILE_NOT_FOUND</code>: error while opening the GML file.
  *   </li>
  *   <li>
  *     <code>GML_NODE_ID_DOUBLE</code>: a node id occures twice.
  *   </li>
  *   <li>
  *     <code>GML_NODE_NO_ID</code>: a node without id.
  *   </li>
  *   <li>
  *     <code>GML_NO_GRAPH_FOUND</code> no keyword graph in the file
  *   </li>
  *   <li>
  *     <code>GML_OK</code>: no error :-))
  *   </li>
  *   <li>
  *     <code>GML_OPEN_BRACKET</code>: too many opening brackets <code>[</code>
  *   </li>
  *   <li>
  *     <code>GML_PREMATURE_EOF</code>: an EOF occured, where it was not expected,
  *       i.e. while parsing a string 
  *   </li>
  *   <li>
  *     <code>GML_SYNTAX</code>: the file is not a valid GML file, i.e. mismatch
  *       in the key-value pairs.
  *   </li>
  *   <li>
  *     <code>GML_TOO_MANY_BRACKETS</code>: a mismatch of brackets was detected,
  *       i.e. too many closing brackets <code>]</code>
  *   </li>
  *   <li>
  *     <code>GML_TOO_MANY_DIGITS</code> the number of digits a floating point or
  *       integer value can hold is limited to 1024
  *   </li>
  *   <li>
  *     <code>GML_UNEXPECTED</code>: a character occured, where i makes no sense,
  *       e.g. non-numerical characters in numbers or keys beginning with numbers 
  *   </li>
  * </ul>
  *
  */
typedef enum
{
    GML_OK=0,
    GML_UNEXPECTED, GML_SYNTAX, GML_PREMATURE_EOF, GML_TOO_MANY_DIGITS,
    GML_OPEN_BRACKET, GML_TOO_MANY_BRACKETS, GML_FILE_NOT_FOUND, GML_NODE_ID_DOUBLE,
    GML_NO_GRAPH_FOUND, GML_NODE_NO_ID, GML_EDGE_NODE_UNKNOWN, GML_EDGE_TWICE
} GML_error_value;


/**
  * @short REason and position of an error in a GML file.
  *
  * When an error accurs while parsing the structure of a GML file
  * <code>GML_error</code> is used to return the type and position of 
  * the error detected. Position is specified by <code>line</code> and
  * <code>column</code>, but might be somewhat imprecise. However at 
  * least the line number hsould not differ too much from the real
  * position.
  *
  * @see graph#loadGML 
  */
struct GML_error
{
    GML_error_value err_num;
    int line;
    int column;
};


/**
  * @internal
  */
union GML_tok_val
{
    long integer;
    double floating;
    char* str;
    struct GML_error err;
};


/**
  * @internal
  */
struct GML_token
{ 
    GML_value kind;
    union GML_tok_val value;
};

/*
 * global variables
 */

extern unsigned int GML_line;
extern unsigned int GML_column;
extern string GML_table[];

/**
  * if you are interested in the position where an error occured it is a good
  * idea to set GML_line and GML_column back. 
  * This is what GML_init does.
  */
 
void GML_init ();

/**
  * returns the error message of an error
  */
string GML_error2string(struct GML_error);

/**
  * returns the next token in file. If an error occured it will be stored in 
  * GML_token.
  */
template <class X> struct GML_token GML_scanner (X source);
     /*     // moved by gil from .cpp to .h 090702
{
    unsigned int cur_max_size = INITIAL_SIZE;
    static char buffer[INITIAL_SIZE];
    char* tmp = buffer;
    char* ret = tmp;
    struct GML_token token;
    int is_float = 0;
    unsigned int count = 0;
    int next;
    char ISO_buffer[8];
    int ISO_count;

    assert (source != NULL);
    
    int eof = seteof(source);

    // eliminate preceeding white spaces
    do
    {
      next = fgetc_ (source);
      GML_column++;

      if (next == '\n')
      {
        GML_line++;
        GML_column = 1;
     	}
        	
    } while (isspace (next) && next != eof);

    // reached EOF
    if (next == eof)
    {
      token.kind = GML_END;
      return token;
    }

    // floating point or integer
    else if (isdigit (next) || next == '.' || next == '+' || next == '-')
    {
      do
      {
        if (count == INITIAL_SIZE - 1)
        {
          token.value.err.err_num = GML_TOO_MANY_DIGITS;
          token.value.err.line = GML_line;
          token.value.err.column = GML_column + count;
          token.kind = GML_ERROR;
        	return token;
        }

        if (next == '.' || next == 'E')
          is_float = 1;

        buffer[count] = next;
        count++;
        next = fgetc_ (source);

      } while (!isspace(next) && next != ']' && next != eof);

      buffer[count] = 0;

      if (next == ']')
	  ungetc_ (next, source); 

      if (next == '\n')
      {
        GML_line++;
        GML_column = 1;
      }
      else
        GML_column += count;
        	
      if (is_float)
      {
        token.value.floating = atof (tmp);
        token.kind = GML_DOUBLE;
      }
      else
      {
        token.value.integer = atol (tmp);
        token.kind = GML_INT;
      }
      return token;
    }

    // key
    else if (isalpha (next) || next == '_')
    {
      do
      {
        if (count == cur_max_size - 1)
        {
          *tmp = 0;
        	tmp =  (char*) malloc(2 * cur_max_size * sizeof (char));
        	strcpy (tmp, ret);
        		
        	if (cur_max_size > INITIAL_SIZE)
        	  free (ret);

       		ret = tmp;
       		tmp += count;
       		cur_max_size *= 2;
       	}
        	
        *tmp++ = next;
        count++;
        next = fgetc_ (source);
      } while (isalnum (next) || next == '_');

      if (next == '\n')
      {
        GML_line++;
        GML_column = 1;
      }
      else
        GML_column += count;
        	
      if (next == '[')
        ungetc_ (next, source);

      else if (!isspace (next))
      {
        token.value.err.err_num = GML_UNEXPECTED;
        token.value.err.line = GML_line;
        token.value.err.column = GML_column + count;
        token.kind = GML_ERROR;
        	
        if (cur_max_size > INITIAL_SIZE)
          free (ret);

        return token;
      }

      *tmp = 0;
      token.kind = GML_KEY;
      token.value.str = (char*) malloc((count+1) * sizeof (char));
      strcpy (token.value.str, ret);
        	
      if (cur_max_size > INITIAL_SIZE)
        free (ret);

      return token;
    }

    //  comments, brackets and strings
    else
    {
      switch (next)
      {
      case '#':
        do
      		next = fgetc_ (source);
        while (next != '\n' && next != EOF);
        		
      	GML_line++;
      	GML_column = 1;
      	return GML_scanner (source);

      case '[':
        token.kind = GML_L_BRACKET;
      	return token;

      case ']':
        token.kind = GML_R_BRACKET;
      	return token;
        		
      case '"':
        next = fgetc_ (source);
      	GML_column++;

      	while (next != '"')
      	{
      		if (count >= cur_max_size - 8)
      		{
      		  *tmp = 0;
      		  tmp = (char*) malloc (2 * cur_max_size * sizeof(char));
      		  strcpy (tmp, ret);
        		
      		  if (cur_max_size > INITIAL_SIZE)
      		    free (ret);

      		  ret = tmp;
      		  tmp += count;
      		  cur_max_size *= 2;
      		}
        	
      		if (next == '&')
      		{
      		  ISO_count = 0;

      		  while (next != ';')
      		  {
      		    if (next == '"' || next == eof)
      		    {
      			    ungetc_ (next, source);
      			    ISO_count = 0;
      			    break;
      			  }
        			
        			if (ISO_count < 8)
        			{
        			  ISO_buffer[ISO_count] = next;
        			  ISO_count++;
        			}

        			next = fgetc_ (source);
        		}
          			
        		if (ISO_count == 8)
        			ISO_count = 0;

        		if (ISO_count)
        		{
        			ISO_buffer[ISO_count] = ';';
        			ISO_count++;
        			next = GML_search_ISO (ISO_buffer, ISO_count);
        			ISO_count = 0;
        		}
          		
        		else
        			next = '&';
        }
      	*tmp++ = next;
      	count++;
      	GML_column++;
        			
      	next = fgetc_ (source);

      	if (next == eof)
      	{
      	  token.value.err.err_num = GML_PREMATURE_EOF;
      		token.value.err.line = GML_line;
      		token.value.err.column = GML_column + count;
      		token.kind = GML_ERROR;
        		
      		if (cur_max_size > INITIAL_SIZE)
      		  free (ret);
        			
   		    return token;
     		}

     		if (next == '\n')
     	  {
      	  GML_line++;
      		GML_column = 1;
      	}
      }

      *tmp = 0;
      token.kind = GML_STRING;
      token.value.str = (char*) malloc((count+1) * sizeof (char));
      strcpy (token.value.str, ret);
        	
      if (cur_max_size > INITIAL_SIZE)
        free (ret);

      return token;
        	
    default:
      token.value.err.err_num = GML_UNEXPECTED;
      token.value.err.line = GML_line;
      token.value.err.column = GML_column;
      token.kind = GML_ERROR;
      return token;
    }		
  }
}
*/

} // namespace graphlib

#endif  // GML_SCANNER_H

/* end of file */



