/***************************************************************************
                          graph.cpp  -  description
                             -------------------
    begin                : Thu Mar 1 2001
    copyright            : (C) 2001 by Gleiss Petra Manuela
    email                : petra@gleiss.at
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "graph.h"

#include "nodedata.h"
#include "nodeid.h"

#include "edgedata.h"
#include "edgeid.h"
#include "adjacencymatrix.h"

#include "utilities.h"
#include "trace.h"
#include "doubleerror.h"

#include <stdio.h>

#include <algorithm>
#include <sstream>

namespace graphlib {
namespace graphlib_utils{ //added by gil 10/05/02 start:
  //using namespace graphlib::graphlib_utils;
  inline string* fopen_( string fname, char* perm)
  {
    return (string*) fopen(fname.c_str(), perm);
  }
  
  inline char** fopen_( char* fname, char* perm)
  {
    return (char**) fopen(fname, perm);
  }

  inline void fclose_( string* file)
  {
    fclose((FILE*) file);
  }

  inline void fclose_( char** file)
  {
    fclose((FILE*) file);
  }

} //added by gil 10/05/02 end.
}
namespace graphlib {
using namespace graphlib::graphlib_utils;
//---------------------------------------------------------------------------
// Con-/Destructors
//---------------------------------------------------------------------------
Graph::Graph()
{
  init();
}

Graph::Graph(const AdjacencyMatrix& am)
{
  init();

  unsigned size=am.size();
  const double *matrix = am.adjacency();

  if (!matrix)
    THROW(runtime_error,"Graph::Graph(const AdjacencyMatrix&)\
      AdjacencyMatrix == NULL")

  for (unsigned i=0; i+1!=size; i++)
  {
    // new node with id=i+1
    newNode();
    for (unsigned j=i; j!=size; j++)
    {
      // 1 -> edge(i+1,j+1)
      if (matrix[i*size+j] == 1)
        newEdge(i+1,j+1);
     }
  }

}

Graph::Graph(const Node& node)
{
  init();
  copyNode(node);
}

Graph::Graph(const Graph& g)
{
  copy(g);
}

Graph::Graph(const Graph& g1, const Graph& g2, unsigned* nb_nodes1)
{
  copy(g1);
  copy2(g2);
  *nb_nodes1 = *g1.freeNodeNumbers.begin();
}

Graph::~Graph()
{
  clear();
}

//---------------------------------------------------------------------------
// Operator
//---------------------------------------------------------------------------
Graph& Graph::operator=(const Graph& g)
{
  copy(g);
  return *this;
}

//---------------------------------------------------------------------------
// Input / Output
//---------------------------------------------------------------------------
template <class X> GML_error Graph::load(const X& fname)
{
  GML_stat stat;
  stat.key_list = NULL;
  GML_pair* keyList;
  GML_pair* origList;

  // initialize stat
  stat.err.line = stat.err.column = 0;

  // open file - do nothing for handle 
  X* file = fopen_(fname, "r"); //changed by gil 10/05/02

  if (!file)
  {
    stat.err.err_num = GML_FILE_NOT_FOUND;
    return stat.err;
  }

  // initialize GML parser, parse file and close it
  GML_init();
  keyList = GML_parser(file, &stat, 0);
  fclose_(file); //changed by gil 10/05/02

  // print keyList
  #ifdef GRAPHLIB_DEBUG
    #if GRAPHLIB_DEBUG>4
    if (logfile.isInit())
      GML_printList(keyList,0,logfile.os());
    #endif
  #endif

  // error occured
  if (stat.err.err_num != GML_OK)
  {
    GML_free_list(keyList, stat.key_list);
    return stat.err;
  }

  //
  // valid GML file -> build the graph
  //
  clear();
  origList = keyList;

  //
  // get the first entry with key "graph" in the list and move it to the
  // beginning
  //
  GML_pair* prev = 0;

  while (keyList)
  {
    if (!strcmp("graph",keyList->key))
      break;

    prev = keyList;
    keyList = keyList->next;
  }

  if (prev)
  {
    prev->next = keyList->next;
    keyList->next = origList;
    origList = keyList;
  }

  // keys of the graph (directed, name, node, edge, ...)
  keyList = keyList->value.list;

  GML_pair* lastNode = keyList;

  GML_pair* tmpList;
  GML_pair* tmpPrev;
  GML_pair* nodeEntries = 0;
  GML_pair* edgeEntries = 0;

  prev = 0;
  bool targetFound, sourceFound;

  //
  // Node and edge keys maybe in arbitrary order -> sort them
  // all nodes before all edges
  //
  while (keyList)
  {
    if (!strcmp(keyList->key, "node"))
    {
      // search the list associated with this node for the id
      // move id to the front of the list
      tmpList = keyList->value.list;
      tmpPrev = 0;

      while (tmpList)
      {
        if (!strcmp(tmpList->key, "id"))
        {
          if (tmpList->kind != GML_INT)
            THROW(runtime_error,"Graph::load(const string&) node id not an integer!")

          if (tmpPrev == 0)
            break;
          else
          {
            tmpPrev->next = tmpList->next;
            tmpList->next = keyList->value.list;
            keyList->value.list = tmpList;
            break;
          }
        }
        tmpPrev = tmpList;
        tmpList = tmpList->next;
      }

      if (prev)
      {
        prev->next = keyList->next;
        keyList->next = nodeEntries;
        nodeEntries = keyList;
        keyList = prev->next;
        continue;
      }
      else
      {
        origList->value.list = keyList->next;
        keyList->next = nodeEntries;
        nodeEntries = keyList;
        keyList = origList->value.list;
        continue;
      }
    }
    else if (!strcmp(keyList->key, "edge"))
    {
      // search for source and target
      // move them to front of list
      tmpList = keyList->value.list;
      tmpPrev = 0;
      sourceFound = false;
      targetFound = false;

      while (tmpList)
      {
        if (!strcmp(tmpList->key, "source"))
        {
          if (tmpList->kind != GML_INT)
            THROW(runtime_error,"Graph::load(const string&) edge source not an integer!")
          sourceFound = true;

          if (tmpPrev)
          {
            tmpPrev->next = tmpList->next;
            tmpList->next = keyList->value.list;
            keyList->value.list = tmpList;
            tmpList = tmpPrev;
            if (targetFound)
              break;
          }

        }
        else if (!strcmp(tmpList->key, "target"))
        {
          if (tmpList->kind != GML_INT)
            THROW(runtime_error,"Graph::load(const string&) edge target not an integer!")
          targetFound = true;

          if (tmpPrev)
          {
            tmpPrev->next = tmpList->next;

            if (sourceFound)
            {
              tmpList->next = keyList->value.list->next;
              keyList->value.list->next = tmpList;
              break;
            }
            else
            {
              tmpList->next = keyList->value.list;
              keyList->value.list->next = tmpList;
              tmpList = tmpPrev;
            }
          }
        }
        tmpPrev = tmpList;
        tmpList = tmpList->next;
      }

      if (prev)
      {
        prev->next = keyList->next;
        keyList->next = edgeEntries;
        edgeEntries = keyList;
        keyList = prev->next;
        continue;
      }
      else
      {
        origList->value.list = keyList->next;
        keyList->next = edgeEntries;
        edgeEntries = keyList;
        keyList = origList->value.list;
        continue;
      }
    }
    else if (!strcmp(keyList->key, "directed"))
    {
      // ignore at the moment
      // remove from key list
      if (prev)
      {
        prev->next = keyList->next;
        free (keyList);
        keyList = prev;
      }
      else
      {
        origList->value.list = keyList->next;
        lastNode = keyList->next;
        free (keyList);
        keyList = origList->value.list;
        continue;
      }
    }
    else if (!strcmp(keyList->key, "name"))
    {
      // name of the graph
      // remove from key list
      //name_ = keyList->value.str;
      if (prev)
      {
        prev->next = keyList->next;
        free (keyList);
        keyList = prev;
      }
      else
      {
        origList->value.list = keyList->next;
        lastNode = keyList->next;
        free (keyList);
        keyList = origList->value.list;
        continue;
      }
    }
    prev = keyList;
    keyList = keyList->next;
  }

  //
  // make the graph described in the list
  //

  // nodes
  loadNodes(nodeEntries, &stat);
  if (stat.err.err_num != GML_OK)
    return stat.err;

  // edges
  loadEdges(edgeEntries, &stat);
  if (stat.err.err_num != GML_OK)
    return stat.err;

  GML_free_list(origList, stat.key_list);
  GML_free_list(nodeEntries,0);
  GML_free_list(edgeEntries,0);
  stat.err.err_num = GML_OK;
  return stat.err;
}

template GML_error Graph::load<string>(const string&); //needed for template linking
template GML_error Graph::load<char*>(char* const&); //needed for template linking
//NOTE! the order char* const& is INTENDED! const char*& will NOT work!


GML_error Graph::loadNodes(GML_pair* nodeEntries, GML_stat* stat)
{
  GML_pair* node;
  GML_pair* graphics;
  double tmp;
  bool nodeHasId=false;

  while (nodeEntries)
  {
    node = nodeEntries->value.list;
    Node n;
    n.data = new NodeData;

    while (node)
    {
      if (!strcmp(node->key, "id"))
      {
        if (getFreeNodeNumber(node->value.integer) != true)
        {
          stat->err.err_num = GML_NODE_ID_DOUBLE;
          stat->err.line = nodeEntries->lineNr;
          return stat->err;
        }
        n.data->id = node->value.integer;
        nodeHasId = true;
      }
      else if (!strcmp(node->key, "label") || !strcmp(node->key, "parameter"))
        n.data->label = node->value.str;

      else if (!strcmp(node->key, "graphics"))
      {
        graphics = node->value.list;

        while (graphics)
        {
          tmp = graphics->value.floating;
          if (!strcmp(graphics->key, "x"))
          {
            n.data->coord.x = tmp;
            //if (maxCoords.x < tmp) maxCoords.x=tmp;
            //else if (minCoords.x > tmp) minCoords.x=tmp;
          }
          else if (!strcmp(graphics->key, "y"))
          {
            n.data->coord.y = tmp;
            //if (maxCoords.y < tmp) maxCoords.y=tmp;
            //else if (minCoords.y > tmp) minCoords.y=tmp;
          }
          else if (!strcmp(graphics->key, "z"))
          {
            n.data->coord.z = tmp;
            //if (maxCoords.z < tmp) maxCoords.z=tmp;
            //else if (minCoords.z > tmp) minCoords.z=tmp;
          }
          graphics = graphics->next;
        }
      }
     node = node->next;
    }
    //node without id
    if (nodeHasId == false)
    {
      stat->err.err_num = GML_NODE_NO_ID;
      stat->err.line = nodeEntries->lineNr;
      delete n.data;
      return stat->err;
    }

    // add new node to graph
    n.data->pos = nodes.insert(nodes.end(),n);
    ++nodes_count;

    // first node
    //if (nodes_count == 1)
    //minCoords = maxCoords = n.coord();

    // next node
    nodeEntries = nodeEntries->next;
  }
  return stat->err;
}

GML_error Graph::loadEdges(GML_pair* edgeEntries, GML_stat* stat)
{
  GML_pair* edge;
  int source, target;
  string label;

  while (edgeEntries)
  {
    edge = edgeEntries->value.list;
    source = target = 0;
    label = "";

    while (edge)
    {
      if (!strcmp(edge->key, "source"))
        source = edge->value.integer;

      else if (!strcmp(edge->key, "target"))
        target = edge->value.integer;

      else if (!strcmp(edge->key, "label") || !strcmp(edge->key, "parameter"))
        label = edge->value.str;

      else
        TRACE("edge: " + Ttos(source) + "->" + Ttos(target) + " unexpected keys:" +
        edge->key)

      edge = edge->next;
    }
    try
    {
      newEdge(source, target, label);
    }
    catch (DoubleError &e)
    {
      THROW(runtime_error,Ttos(edgeEntries->lineNr)+e.what())
      stat->err.err_num = GML_EDGE_NODE_UNKNOWN;
      stat->err.line = edgeEntries->lineNr;
      return stat->err;
    }
    catch (logic_error &e)
    {
      stat->err.err_num = GML_EDGE_NODE_UNKNOWN;
      stat->err.line = edgeEntries->lineNr;
      return stat->err;
    }
    edgeEntries = edgeEntries->next;
  }
  return stat->err;
}



ostream& operator<< (ostream& os, const Graph& g)
{
  constNodeIter i_n;
  constEdgeIter i_e;

  os.precision(1);
  os.setf(ios::fixed,ios::floatfield);

  os << "# gml undirected Graph" << endl;
  os << "# " << timeStamp() << endl;
  os << endl << "graph [" << endl;
  if (g.name!="") os << "  name \"" << g.name << "\"" << endl;

  if (!g.nodes.empty())
  {
    for (i_n=g.nodes.begin();i_n!=g.nodes.end();i_n++)
    {
      os << "  node [\n    id " << i_n->id();
      os << "\n    label \"" << i_n->label() << "\"";
      os << "\n    graphics [" << i_n->coord();
      os << "\n    ]";
      os << "\n  ]" << endl;
    }
  }

  if (!g.edges.empty())
  {
    for(i_e=g.edges.begin();i_e!=g.edges.end();i_e++)
    {
        os << "  edge [\n    source " << i_e->source().id();
        os << "\n    target " << i_e->target().id() ;
        if (i_e->label() != "")
          os << "\n    label \"" << i_e->label() << "\"";
        os << "\n  ]" << endl;
    }
  }

  os << "]" << endl;
  return os;

}

//---------------------------------------------------------------------------
// Information
//---------------------------------------------------------------------------
unsigned Graph::numberNodes() const
{
  return nodes_count;
}
unsigned Graph::numberEdges() const
{
  return  edges_count;
}

//---------------------------------------------------------------------------
// Search
//---------------------------------------------------------------------------
constNodeIter Graph::findNode(unsigned id) const  	
{
  return find_if(nodes.begin(),nodes.end(),NodeId(id));
}
NodeIter Graph::findNode(unsigned id)  	
{
  return find_if(nodes.begin(),nodes.end(),NodeId(id));
}

constEdgeIter Graph::findEdge(unsigned s, unsigned t) const  	
{
  return find_if(edges.begin(),edges.end(),EdgeSourceTarget(s,t));
}
EdgeIter Graph::findEdge(unsigned s, unsigned t)  	
{
  return find_if(edges.begin(),edges.end(),EdgeSourceTarget(s,t));
}

constEdgeIter Graph::findEdge(unsigned id) const  	
{
  return find_if(edges.begin(),edges.end(),EdgeId(id));
}
EdgeIter Graph::findEdge(unsigned id)   	
{
  return find_if(edges.begin(),edges.end(),EdgeId(id));
}

//---------------------------------------------------------------------------
// Creation
//---------------------------------------------------------------------------
NodeIter Graph::newNode()
{
  // create node
  Node n;
  n.data = new NodeData;
  n.data->id = getFreeNodeNumber();
  n.data->pos=nodes.insert(nodes.end(),n);
  ++nodes_count;

  return n.data->pos;
}

EdgeIter Graph::newEdge(const Node& s, const Node& t, string label)
{
  string error="Graph::newEdge(const Node&, const Node&, string): ";

  if (!s.data || !t.data)
    THROW(logic_error,(error+"source or target has no data."))

  if (findEdge(s.id(),t.id()) != edges.end())
    return findEdge(s.id(),t.id());
    //THROW(DoubleError,(error+"edge already exists."))

  // create edge
  Edge e;
  e.data = new EdgeData;

  // set id and source and target
  e.data->id = getFreeEdgeNumber();
  e.data->nodes.push_back(s);
  e.data->nodes.push_back(t);
  e.data->label=label;

  // add edge to graph
  e.data->pos=edges.insert(edges.end(),e);
  ++edges_count;

  // set adj_pos
  list<Edge>& sourceAdj = s.data->edges;
  list<Edge>& targetAdj = t.data->edges;

  e.data->adj_pos.push_front(sourceAdj.insert(sourceAdj.end(),e));
  e.data->adj_pos.push_back(targetAdj.insert(targetAdj.begin(),e));

  s.data->degree++;
  t.data->degree++;

  return e.data->pos;
}

EdgeIter Graph::newEdge(unsigned source, unsigned target, string label)
{
  string error="Graph::newEdge(unsigned,unsigned,string): ";
  ostringstream err;
  err << "Edge (" << source << "," << target << ") ";
  error += err.str();

  constNodeIter s,t;
  if ((s=findNode(source)) == nodes.end() ||
      (t=findNode(target)) == nodes.end())
    THROW(logic_error,error+"source or target not found")

  return newEdge(*s,*t,label);
}

Graph* Graph::makeSubgraph(cVNodeIter begin, cVNodeIter end)
{
  // removes all nodes and edges
  clear();

  cVNodeIter it = begin;

  while (it != end)
  {
    // copy node
    copyNode(*it);

    // copy adjacent edges
    constEdgeIter edge;
    for (edge = it->adjBegin(); edge!=it->adjEnd(); edge++)
    {
      if (findNode(it->opposite(*edge).id()) != nodes.end())
        copyEdge(*edge);
    }
    it++;
  }
  return this;
}

Graph* Graph::makeSubgraph(cVEdgeIter begin, cVEdgeIter end)
{
  // removes all nodes and edges
  clear();

  cVEdgeIter it = begin;

  while (it != end)
  {
    // copy source and target of edge
    try
    {
      copyNode(it->source());
    }
    catch (DoubleError)  { ; } // node already exists

    try
    {
      copyNode(it->target());

    }
    catch (DoubleError) { ; } // node already exists

    copyEdge(*it);
    it++;
  }


  return this;
}

//---------------------------------------------------------------------------
// Deletion
//---------------------------------------------------------------------------
void Graph::delNode(Node& n)
{
  if (!n.data )
    THROW(logic_error,"Graph::delNode(const Node&): node has no data.")
  unsigned freeNumber = n.data->id;

  // delete incident edges
  while (n.data->edges.begin() != n.data->edges.end())
  {
    delEdge(*n.data->edges.begin());
  }

  // delete node from the graph
  nodes.erase(n.data->pos);
  --nodes_count;
  setFreeNodeNumber(freeNumber);
  //delete n.data; n.data=NULL; //change by gil 050802
}

void Graph::delAllNodes()
{
  clear();
}

void Graph::delEdge(Edge& e)
{
  unsigned freeNumber = e.data->id;
  if (!e.data)
    THROW(logic_error,"Graph::delEdge(const Edge&): edge has no data.")

  // remove edge form adjacency list
  e.removeFrom();

  // delete edge from the graph
  edges.erase(e.data->pos);
  --edges_count;
  setFreeEdgeNumber(freeNumber);
  //delete e.data; e.data=NULL;
}

void Graph::delAllEdges()
{
  constNodeIter it = nodes.begin();
  constNodeIter end = nodes.end();

  while (it != end)
  {
    (it++)->data->edges.clear();
  }

  delList(edges);

  edges_count = 0; setFreeNodeNumber(0);

}

void Graph::clear()
{
  delList(edges);
  delList(nodes);

  nodes_count = edges_count = 0;
  setFreeNodeNumber(0);
  setFreeEdgeNumber(0);
}

//---------------------------------------------------------------------------
// Utilities for copy() - CopyConstructor and AssignementOperator
//---------------------------------------------------------------------------
void Graph::init()
{
  nodes_count=0;
  edges_count=0;
  setFreeNodeNumber(0);
  setFreeEdgeNumber(0);
  nodes.clear();
  edges.clear();
}

void Graph::copy(const Graph& g)
{
  // clear all nodes and edges
  clear();

  // copy all nodes
  copyList(g.nodesBegin(),g.nodesEnd());

  // copy all edges
  copyList(g.edgesBegin(),g.edgesEnd());

  // copy other attributes
  name = g.name;
  nodes_count = g.nodes_count;
  edges_count = g.edges_count;
  freeNodeNumbers.clear();
  freeNodeNumbers.insert(freeNodeNumbers.end(),
                         g.freeNodeNumbers.begin(),g.freeNodeNumbers.end());
  freeEdgeNumbers.clear();
  freeEdgeNumbers.insert(freeEdgeNumbers.end(),
                         g.freeEdgeNumbers.begin(),g.freeEdgeNumbers.end());

}


void Graph::copy2(const Graph& g)
{
  int nb_nodes1 = *freeNodeNumbers.begin(); //numberNodes();
  {
  // copy all nodes
  constNodeIter old = g.nodesBegin(); 
  constNodeIter end = g.nodesEnd();
  while(old != end)
  {
    if (findNode(old->id()+nb_nodes1) != nodes.end())
    {
      ostringstream err;
      err << "Graph::copyNode(const Node&): id (" << old->id()+nb_nodes1
          << ") already exists." << ends;
      throw(DoubleError(err.str()));
    }

    // create node
    Node n;
    n.data = new NodeData;
    
    // set id, label, ...
    n.data->id     = old->id()+nb_nodes1;
    n.data->label  = old->label();
    n.data->coord  = old->coord();
    n.data->degree = 0;
    
    // add to graph
    n.data->pos=nodes.insert(nodes.end(),n);
    ++nodes_count;
    old++;
  }
  }

  {
  // copy all edges
  constEdgeIter old = g.edgesBegin(); 
  constEdgeIter end = g.edgesEnd();
  while(old != end)
  {
    newEdge(old->source().id()+nb_nodes1,old->target().id()+nb_nodes1,
            old->label());
    old++;
  }
  }
    freeNodeNumbers.clear();
    freeNodeNumbers.push_back(nb_nodes1);
}


void Graph::copyList(constNodeIter it, constNodeIter end)
{
  while(it != end)
  {
    copyNode(*(it++));
  }
}

void Graph::copyNode(const Node& old)
{
  if (findNode(old.id()) != nodes.end())
  {
    ostringstream err;
    err << "Graph::copyNode(const Node&): id (" << old.id()
        << ") already exists.";// << ends;
    throw(DoubleError(err.str()));
  }

  // create node
  Node n;
  n.data = new NodeData;

  // set id, label, ...
  n.data->id     = old.id();
  n.data->label  = old.label();
  n.data->coord  = old.coord();
  n.data->degree = 0;

  // add to graph
  n.data->pos=nodes.insert(nodes.end(),n);
  ++nodes_count;
}

void Graph::copyList(constEdgeIter it, constEdgeIter end)
{
  while(it != end)
  {
    copyEdge(*(it++));
  }
}

void Graph::copyEdge(const Edge& old)
{
  newEdge(old.source().id(),old.target().id(),old.label());
}

//---------------------------------------------------------------------------
// Utilities for clear() - Destructor
//---------------------------------------------------------------------------
void Graph::delList(list<Node>& l)
{
  constNodeIter it = l.begin();
  constNodeIter end = l.end();

  while(it != end)
  {
    delete (it++)->data;
  }
  l.clear();
}

void Graph::delList(list<Edge>& l)
{
  constEdgeIter it = l.begin();
  constEdgeIter end = l.end();

  while(it != end)
  {
    delete (it++)->data;
  }
  l.clear();
}

//---------------------------------------------------------------------------
// Utilities for creating and deleting nodes/edges
//---------------------------------------------------------------------------
bool Graph::getFreeNodeNumber(unsigned number)
{
  unsigned tmp;

  // number is bigger then the first list element
  tmp = *freeNodeNumbers.begin();
  if (number > tmp)
  {
    // new first list element
    *freeNodeNumbers.begin() = number;
    // add all numbers between number and the old first list element
    list<unsigned>::iterator it = ++(freeNodeNumbers.begin());
    for (unsigned i=number-1; i!=tmp; i--)
      freeNodeNumbers.insert(it,i);

    return true;
  }

  else
  {
    list<unsigned>::iterator it;
    // search for the number in the list
    it = find(freeNodeNumbers.begin(),freeNodeNumbers.end(),number);

    // number is in the list
    if (it != freeNodeNumbers.end())
    {
      freeNodeNumbers.erase(it);
      return true;
    }

    // number is not in the list
    else
       return false;

  }
}

unsigned Graph::getFreeNodeNumber()
{
  unsigned tmp;
  if (freeNodeNumbers.size() == 1)
    tmp = ++(*freeNodeNumbers.begin());

  else
  {
    tmp = freeNodeNumbers.back();
    freeNodeNumbers.pop_back();
  }
  return tmp;
}

void Graph::setFreeNodeNumber(unsigned number)
{
  // first element
  if (number == 0)
  {
    freeNodeNumbers.clear();
    freeNodeNumbers.push_back(0);
    return;
  }

  // node number is free -> add
  else
  {
    freeNodeNumbers.insert(++freeNodeNumbers.begin(),number);
    return;
  }
}

unsigned Graph::getFreeEdgeNumber()
{
  unsigned tmp;
  if (freeEdgeNumbers.size() == 1)
    tmp = ++(*freeEdgeNumbers.begin());

  else
  {
    tmp = freeEdgeNumbers.back();
    freeEdgeNumbers.pop_back();
  }

 return tmp;
}

void Graph::setFreeEdgeNumber(unsigned number)
{
  // first element
  if (number == 0)
  {
    freeEdgeNumbers.clear();
    freeEdgeNumbers.push_back(0);
    return;
  }

  // Edge number is free -> add
  else
  {
    freeEdgeNumbers.insert(++freeEdgeNumbers.begin(),number);
    return;
  }
}


} // namespace graphlib

/* end of file */
