/***************************************************************************
                          graph.h  -  description
                             -------------------
    begin                : Thu Mar 1 2001
    copyright            : (C) 2001 by Gleiss Petra Manuela
    email                : petra@gleiss.at
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef GRAPH_H
#define GRAPH_H

#include "node.h"
#include "edge.h"

#include "gml_parser.h"

#include <string>
#include <vector>
#include <iostream>
#include <list>

#include "trace.h"
#include "logfile.h"
#include "doubleerror.h"

/*
namespace { //added by gil 10/05/02 start:
  using namespace graphlib::graphlib_utils;
  string* fopen_( string fname, char* perm)
  {
    return (string*) fopen(fname.c_str(), perm);
  }
  
  char** fopen_( char* fname, char* perm)
  {
    return (char**) fopen(fname, perm);
  }

  void fclose_( string* file)
  {
    fclose((FILE*) file);
  }

  void fclose_( char** file)
  {
    fclose((FILE*) file);
  }

} //added by gil 10/05/02 end.
*/
namespace graphlib {

typedef vector<Node>::const_iterator cVNodeIter;
typedef vector<Edge>::const_iterator cVEdgeIter;
typedef list<Graph*>::iterator             GraphIter;
typedef list<Graph*>::const_iterator       constGraphIter;

class AdjacencyMatrix;
/**
  *@short An undirected graph
  *
  * A graph <var>G=(V,E)</var> consists of a set of nodes <var>V</var>
  * and a set of edges <var>E</var>, where every edge can be viewed as
  * an unordered pair of nodes <var>(s,t)</var> connecting source
  * <var>s</var> with target <var>t</var>.
  *
  * @see Edge
  * @see Node
  *
  */

class Graph
{
  	//------------------------------------------------------ friends
    /**
      * Output operator.
      * Creates an GML output of the <code>g</code>.
      *
      * @param os the output-stream
      * @param g the graph
      * @see <a href="http://www.infosun.fmi.uni-passau.de/Graphlet/
      *       GML/index.html">The GML File Format</a>
      * @return the output-stream
      */
  friend ostream& operator<< (ostream& os, const Graph& g);

  public:
  	//------------------------------------------------------ Con-/Destructors
    /**
      * Constructor.
      * Generates an empty graph, without any nodes and any edges.
      */
  	Graph();
  	  	
    /**
      * Constructor.
      * Generates a graph from the <code>adjacency matrix</code>.
      *
      * @param AdjacencyMatrix
      */
  	Graph(const AdjacencyMatrix&);
  	  	
    /**
      * Constructor.
      * Generates a graph with the single <code>node</code>.
      *
      * @param Node
      */
  	Graph(const Node&);
  	  	
  	/**
  	  * Copy constructor.
  	  * <em>Please note:</em> This will generate an
  	  * isomorpic copy of <code>g</code>. Althoug this graph will look like
  	  * <code>g</code> it is not <em>physically</em> the same. Especially
  	  * it consists of nodes and edges, which of course have counterparts
  	  * in <code>g</code>, but are different.
  	  *
  	  * @param Graph the original graph
  	  */
    Graph(const Graph& g);  	
  	
        /**
          * Copy constructor.
          * <em>Please note:</em> This will generate an
          * isomorpic copy of <code>g</code>. Althoug this graph will look like
          * <code>g</code> it is not <em>physically</em> the same. Especially
          * it consists of nodes and edges, which of course have counterparts
          * in <code>g</code>, but are different.
          *
          * @param Graph the original graph
          */
        Graph(const Graph& g1, const Graph& g2, unsigned* nb_nodes1);

  	/**
  	  * Destructor. Deletes all nodes and edges.
  	  */
  	virtual ~Graph();
  	
  	//------------------------------------------------------ Operators
  	/**
  	  * Assignment Operator.
  	  * <em>Please note:</em> This will generate an
  	  * isomorpic copy of <code>g</code>. Althoug this graph will look like
  	  * <code>g</code> it is not <em>physically</em> the same.
  	  *
  	  * @param g the original graph
  	  * @see Graph#Graph
  	  */
  	Graph& operator=(const Graph& g);
  	
  	//------------------------------------------------------ Information  	
  	/**
  	  * Returns the number of nodes in the graph.
  	  *
  	  * @return number of nodes
  	  */
  	unsigned numberNodes() const;
  	
  	/**
  	  * Returns the number of edges in the graph.
  	  *
  	  * @return number of edges
  	  */
  	unsigned numberEdges() const;
  	
  	//------------------------------------------------------ Creation
  	/**
  	  * Adds a new node
  	  *
  	  * @param id of node
  	  * @return iterator to new node
  	  */
  	NodeIter newNode();
  	
  	/**
  	  * Adds a new edge from <code>s</code> to <code>t</code>.
  	  * Increases degrees and adds to neighbours from <code>s</code>
  	  * and <code>t</code>.
  	  *
  	  * <p>
  	  * <em>Precondition:</em> <code>s,t</code> are valid nodes in the graph.
  	  *
  	  * @param s source of new edge
  	  * @param t target of new edge
  	  * @param label label of new edge (optional)
  	  *
  	  * @return iterator to new edge
  	  */
    EdgeIter newEdge(const Node& s, const Node& t, string label="");
  	
  	/**
  	  * Adds a new edge from <code>s</code> to <code>t</code> with
  	  * <code>label</code>.
  	  * Increases degrees and adds to neighbours from <code>s</code> and
  	  * <code>t</code>.
  	  * <code>label</code> is optional.
  	  *
  	  * <p>
  	  * <em>Precondition:</em> <code>s,t</code> are valid nodes in the graph.
  	  *
  	  * @param s source of new edge
  	  * @param t target of new edge
  	  * @param label label of new edge (optional)
  	  *
  	  * @return iterator to new edge
  	  */
    EdgeIter newEdge(unsigned source, unsigned target, string label="");

    /**
      * Generates a subgraph with all nodes form <code>begin</code>
      * to <code>end</code> and its adjacent edges.
      * <em>Please note:</em> This will generate isomorpic copies
      * of nodes and edges.
      *
      * @param <code>begin</code> of a vector of node
      * @param <code>end</code> of a vector node
      *
      * @return pointer to itself.
      */
    Graph* makeSubgraph(cVNodeIter begin, cVNodeIter end);

    /**
      * Generates a subgraph with all edges form <code>begin</code>
      * to <code>end</code>.
      * <em>Please note:</em> This will generate isomorpic copies
      * of nodes and edges.
      *
      * @param <code>begin</code> of a vector of edges
      * @param <code>end</code> of a vector edges
      *
      * @return pointer to itself.
      */
    Graph* makeSubgraph(cVEdgeIter begin, cVEdgeIter end);

   	//------------------------------------------------------ Deletion
  	/**
  	  * Deletes node <code>n</code>, and thus all edges incident with
  	  * <code>n</code>.
  	  * <p>
  	  * <em>Precondition:</em> <code>n</code> is a valid nodes in the graph,
  	  * else a logic_error will be thrown.
  	  *
  	  * @param n valid node to be deleted
  	  */
  	void delNode(Node& n);

    /**
  	 	* Deletes all nodes and edges of this graph.
  	 	*
  	 	* @see Graph#clear
  	 	*/
  	void delAllNodes();
  	
  	/**
  	  * Deletes edge <code>e</code>.
  	  * <p>
  	  * <em>Precondition:</em> <code>e</code> is a valid edge in the graph,
  	  * else a logic_error will be thrown.
  	  *
  	  * @param <code>e</code> valid edge to be deleted
  	  */
  	void delEdge(Edge& e);

    /**
  	 	* Deletes all edges of this graph.
  	 	*/
  	void delAllEdges();
  	
  	/**
  	  * Deletes all nodes and edges.
  	  */
  	void clear();
  	
    //------------------------------------------------------ I/O
    /**
      * Load graph from a file in GML-format.
      *
      * @param fname file in GML-format.
      * @return detailed error description (hopefully GML_OK). For
      *   details see @ref GML_error#err_num.
      * @see
      *   <a href="http://www.infosun.fmi.uni-passau.de/Graphlet/GML/index.html">
      *   The GML File Format</a>
      */
    template <class X> GML_error load(const X& fname);
    /* //moved from .cpp to .h by gil 09/07/02
{
  GML_stat stat;
  stat.key_list = NULL;
  GML_pair* keyList;
  GML_pair* origList;

  // initialize stat
  stat.err.line = stat.err.column = 0;

  // open file - do nothing for handle 
  X* file = fopen_(fname, "r"); //changed by gil 10/05/02

  if (!file)
  {
    stat.err.err_num = GML_FILE_NOT_FOUND;
    return stat.err;
  }

  // initialize GML parser, parse file and close it
  GML_init();
  keyList = GML_parser(file, &stat, 0);
  fclose_(file); //changed by gil 10/05/02

  // print keyList
  #ifdef GRAPHLIB_DEBUG
    #if GRAPHLIB_DEBUG>4
    if (logfile.isInit())
      GML_printList(keyList,0,logfile.os());
    #endif
  #endif

  // error occured
  if (stat.err.err_num != GML_OK)
  {
    GML_free_list(keyList, stat.key_list);
    return stat.err;
  }

  //
  // valid GML file -> build the graph
  //
  clear();
  origList = keyList;

  //
  // get the first entry with key "graph" in the list and move it to the
  // beginning
  //
  GML_pair* prev = 0;

  while (keyList)
  {
    if (!strcmp("graph",keyList->key))
      break;

    prev = keyList;
    keyList = keyList->next;
  }

  if (prev)
  {
    prev->next = keyList->next;
    keyList->next = origList;
    origList = keyList;
  }

  // keys of the graph (directed, name, node, edge, ...)
  keyList = keyList->value.list;

  GML_pair* lastNode = keyList;

  GML_pair* tmpList;
  GML_pair* tmpPrev;
  GML_pair* nodeEntries = 0;
  GML_pair* edgeEntries = 0;

  prev = 0;
  bool targetFound, sourceFound;

  //
  // Node and edge keys maybe in arbitrary order -> sort them
  // all nodes before all edges
  //
  while (keyList)
  {
    if (!strcmp(keyList->key, "node"))
    {
      // search the list associated with this node for the id
      // move id to the front of the list
      tmpList = keyList->value.list;
      tmpPrev = 0;

      while (tmpList)
      {
        if (!strcmp(tmpList->key, "id"))
        {
          if (tmpList->kind != GML_INT)
            THROW(runtime_error,"Graph::load(const string&) node id not an integer!")

          if (tmpPrev == 0)
            break;
          else
          {
            tmpPrev->next = tmpList->next;
            tmpList->next = keyList->value.list;
            keyList->value.list = tmpList;
            break;
          }
        }
        tmpPrev = tmpList;
        tmpList = tmpList->next;
      }

      if (prev)
      {
        prev->next = keyList->next;
        keyList->next = nodeEntries;
        nodeEntries = keyList;
        keyList = prev->next;
        continue;
      }
      else
      {
        origList->value.list = keyList->next;
        keyList->next = nodeEntries;
        nodeEntries = keyList;
        keyList = origList->value.list;
        continue;
      }
    }
    else if (!strcmp(keyList->key, "edge"))
    {
      // search for source and target
      // move them to front of list
      tmpList = keyList->value.list;
      tmpPrev = 0;
      sourceFound = false;
      targetFound = false;

      while (tmpList)
      {
        if (!strcmp(tmpList->key, "source"))
        {
          if (tmpList->kind != GML_INT)
            THROW(runtime_error,"Graph::load(const string&) edge source not an integer!")
          sourceFound = true;

          if (tmpPrev)
          {
            tmpPrev->next = tmpList->next;
            tmpList->next = keyList->value.list;
            keyList->value.list = tmpList;
            tmpList = tmpPrev;
            if (targetFound)
              break;
          }

        }
        else if (!strcmp(tmpList->key, "target"))
        {
          if (tmpList->kind != GML_INT)
            THROW(runtime_error,"Graph::load(const string&) edge target not an integer!")
          targetFound = true;

          if (tmpPrev)
          {
            tmpPrev->next = tmpList->next;

            if (sourceFound)
            {
              tmpList->next = keyList->value.list->next;
              keyList->value.list->next = tmpList;
              break;
            }
            else
            {
              tmpList->next = keyList->value.list;
              keyList->value.list->next = tmpList;
              tmpList = tmpPrev;
            }
          }
        }
        tmpPrev = tmpList;
        tmpList = tmpList->next;
      }

      if (prev)
      {
        prev->next = keyList->next;
        keyList->next = edgeEntries;
        edgeEntries = keyList;
        keyList = prev->next;
        continue;
      }
      else
      {
        origList->value.list = keyList->next;
        keyList->next = edgeEntries;
        edgeEntries = keyList;
        keyList = origList->value.list;
        continue;
      }
    }
    else if (!strcmp(keyList->key, "directed"))
    {
      // ignore at the moment
      // remove from key list
      if (prev)
      {
        prev->next = keyList->next;
        free (keyList);
        keyList = prev;
      }
      else
      {
        origList->value.list = keyList->next;
        lastNode = keyList->next;
        free (keyList);
        keyList = origList->value.list;
        continue;
      }
    }
    else if (!strcmp(keyList->key, "name"))
    {
      // name of the graph
      // remove from key list
      //name_ = keyList->value.str;
      if (prev)
      {
        prev->next = keyList->next;
        free (keyList);
        keyList = prev;
      }
      else
      {
        origList->value.list = keyList->next;
        lastNode = keyList->next;
        free (keyList);
        keyList = origList->value.list;
        continue;
      }
    }
    prev = keyList;
    keyList = keyList->next;
  }

  //
  // make the graph described in the list
  //

  // nodes
  loadNodes(nodeEntries, &stat);
  if (stat.err.err_num != GML_OK)
    return stat.err;

  // edges
  loadEdges(edgeEntries, &stat);
  if (stat.err.err_num != GML_OK)
    return stat.err;

  GML_free_list(origList, stat.key_list);
  GML_free_list(nodeEntries,0);
  GML_free_list(edgeEntries,0);
  stat.err.err_num = GML_OK;
  return stat.err;
}
*/

  	//------------------------------------------------------ Iterators  	
  	/**
  	  * Iterate through all nodes in the graph.
  	  *
  	  * @return start for iteration through all nodes in the graph.
  	  */
  	constNodeIter nodesBegin() const { return nodes.begin(); }
  	
  	/**
  	  * Iterate through all nodes in the graph.
  	  *
  	  * @return end for iteration through all nodes in the graph.
  	  */
  	constNodeIter nodesEnd() const { return nodes.end(); }
  	
  	/**
  	  * Iterate through all edges in the graph.
  	  *
  	  * @return start for iteration through all edges in the graph.
  	  */
  	constEdgeIter edgesBegin() const { return edges.begin(); }
  	
  	/**
  	  * Iterate through all edges in the graph.
  	  *
  	  * @return end for iteration through all edges in the graph.
  	  */
  	constEdgeIter edgesEnd() const { return edges.end(); }
  	
  	//------------------------------------------------------ Search
  	/**
  	  * Returns the position in the list of all nodes of node with
  	  * <code>id</code>. If there's no node with <code>id</code>
  	  * the end will be returned.
  	  *
  	  * @param id of a valid node
  	  * @return const_iterator to the node or end
  	  */
  	constNodeIter findNode(unsigned id) const;  	
  	
  	/**
  	  * Returns the position in the list of all nodes of node with
  	  * <code>id</code>. If there's no node with <code>id</code>
  	  * the end will be returned.
  	  *
  	  * @param id of a valid node
  	  * @return iterator to the node or end
  	  */
  	NodeIter findNode(unsigned id);  	
  	
  	/**
  	  * Returns the position in the list of all edges of an edge
  	  * with <code>id</code>. If there's no edge with <code>id</code>
  	  * the end will be returned.
  	  *
  	  * @param <code>id</code> id of a valid edge
  	  * @return const_iterator to the edge or end
  	  */
  	constEdgeIter findEdge(unsigned id) const;  	
  	
  	/**
  	  * Returns the position in the list of all edges of an edge
  	  * with <code>id</code>. If there's no edge with <code>id</code>
  	  * the end will be returned.
  	  *
  	  * @param <code>id</code> id of a valid edge
  	  * @return iterator to the edge or end
  	  */
  	EdgeIter findEdge(unsigned id);  	
  	
  	/**
  	  * Returns the position in the list of all edges of an edge
  	  * with <code>s</code>ource and <code>t</code>arge. If there's no
  	  * edge with <code>s</code> and <code>t</code> the end will
  	  * be returned.
  	  *
  	  * @param <code>s</code> source of a valid edge
  	  * @param <code>t</code> target of a valid edge
  	  * @return const_iterator to the edge or end
  	  */
  	constEdgeIter findEdge(unsigned s, unsigned t) const;  	
  	
  	/**
  	  * Returns the position in the list of all edges of an edge
  	  * with <code>s</code>ource and <code>t</code>arge. If there's no
  	  * edge with <code>s</code> and <code>t</code> the end will
  	  * be returned.
  	  *
  	  * @param <code>s</code> source of a valid edge
  	  * @param <code>t</code> target of a valid edge
  	  * @return iterator to the edge or end
  	  */
  	EdgeIter findEdge(unsigned s, unsigned t);  	
  	
  public:
    //------------------------------------------------------ Name
    /// name of the graph
    string        name;
  protected:
    //------------------------------------------------------ Flag
    /// number of nodes
    unsigned      nodes_count;
    /// number of edges
    unsigned      edges_count;
    /// list of free node numbers - first element = biggest number,
    /// other elements = free numbers between 1 and the biggest one
    list<unsigned> freeNodeNumbers;
    /// list of free edge numbers - first element = biggest number,
    /// other elements = free numbers between 1 and the biggest one
    list<unsigned> freeEdgeNumbers;
  	
  	//------------------------------------------------------ Nodes/Edges
    /// list of nodes
    list<Node>    nodes;
    /// list of edges
    list<Edge>    edges;
  	
  	//------------------------------------------------------ Utilities
    /// Needed for creating isomorphic copies of a graph.
  	void copy(const Graph& g);
    /// Needed for concatenating isomorphic copies of a graph.
        void copy2(const Graph& g);
        	
  private:	
  	//------------------------------------------------------ for copy
    /// Needed for creating isomorphic copies of a graph.
  	void copyList(constNodeIter, constNodeIter);
    /// Needed for creating isomorphic copies of a graph.
  	void copyList(constEdgeIter, constEdgeIter);
    /// Needed for creating isomorphic copies of a graph.
  	void copyNode(const Node& n);
    /// Needed for creating isomorphic copies of a graph.
    void copyEdge(const Edge& e);
  	//------------------------------------------------------ for clear
    /// Needed for deleting nodes
  	void delList(list<Node>&);
    /// Needed for deleting edges
  	void delList(list<Edge>&);
  	
  	//------------------------------------------------------ for creation
  protected:  	
  	/// Needed for creating new nodes
  	bool getFreeNodeNumber(unsigned);
  private:  	
    /// doing the initializing things
    void init();
  	/// Needed for creating new nodes
  	unsigned getFreeNodeNumber();
  	/// Needed for deleting nodes
  	void setFreeNodeNumber(unsigned);
  	/// Needed for creating new edges
  	unsigned getFreeEdgeNumber();
  	/// Needed for deleting edges
  	void setFreeEdgeNumber(unsigned);
    /// for reading graphs from a gml file
    GML_error loadNodes(GML_pair*, GML_stat*);
    /// for reading graphs from a gml file
    GML_error loadEdges(GML_pair*, GML_stat*);
  	
};  // class Graph

} // namespace graphlib
#endif // GRAPH_H

/* end of file */
