/***************************************************************************
                          randomgraph.cpp  -  description
                             -------------------
    begin                : Wed Mar 28 2001
    copyright            : (C) 2001 by Gleiss Petra Manuela
    email                : petra@gleiss.at
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "randomgraph.h"
#include "stdlib.h"

#include "trace.h"
#include "utilities.h"
#include "coordinates.h"

#include <cmath>
#include <set>
#include <vector>
#include <map>

namespace {


unsigned short xsubi[3];

void init_rand(void)
{
  time_t t;
  (void) time(&t);
  xsubi[0] = xsubi[1] = xsubi[2] = (unsigned short) t;  /* lower 16 bit */
  xsubi[1] += (unsigned short) ((unsigned)t >> 6);
  xsubi[2] += (unsigned short) ((unsigned)t >> 12);
#ifndef HAVE_ERAND48
  srand((unsigned int) t); 
#endif
}

double urn()
     /* uniform random number generator; urn() is in [0,1] */
     /* uses a linear congruential library routine */
     /* 48 bit arithmetic */
{
#define HAVE_ERAND 1
#ifdef HAVE_ERAND48
  extern double erand48(unsigned short[]);
  return erand48(xsubi);
#else
  return ((double) rand())/RAND_MAX;
#endif
}

}

namespace graphlib {

//---------------------------------------------------------------------------
// Con-/Destructors
//---------------------------------------------------------------------------
RandomGraph::RandomGraph(unsigned nodes, unsigned edges, 
			 double prob, GraphFlag flag)
       : Graph()
{
  static bool ersteMal = true;
  if (ersteMal) 
  {
    init_rand();
    ersteMal = false;
  }

  SubGraph* tmp=0;

  //  no nodes
  if (nodes==0 && edges==0)
    return;

  if (edges == nodes*(nodes-1)/2)
    flag = Complete;

  if (flag == ScaleFree)
    tmp = makeScaleFree(nodes,edges,prob);

  else if (flag == SmallWorld)
    tmp = makeSmallWorld(nodes,edges, prob);

  else if (flag == ErdoesRenyi)
    tmp = makeErdoesRenyi(nodes,edges);

  else if (flag == Random)
    tmp = makeRandom(nodes,edges);

  else if (flag == Hamiltonian)
    tmp = makeHamilton(nodes,edges);

  else if (flag == Complete)
    tmp = makeComplete(nodes);

  else if (flag == Bipartite)
    tmp = makeBipartite(nodes,edges);

  else
    ;

  if (tmp) delete tmp;
}

RandomGraph::~RandomGraph()
{ ; }

//---------------------------------------------------------------------------
// additional functions
//---------------------------------------------------------------------------
SubGraph* RandomGraph::makeBipartite(unsigned m, unsigned n,
                                      int xm, int ym1, int ym2,
                                      int xn, int yn1, int yn2)
{
  list<Node> firstN, secondN;
  SubGraph* subgraph = new SubGraph(this);

  NodeIter node;
  bool coord;
  int spaceM, spaceN;

  // coordinates
  if (xm==0 && ym1==0 && ym2==0 && xn==0 && ym1==0 && ym2==0)
    coord = false;
  else
  {
    coord = true;
    if (ym1 > ym2)
      swap(ym1,ym2);
    if (yn1 > yn2)
      swap(yn1,yn2);

    spaceM = (m<2) ? 0 : (ym2-ym1)/(m-1);
    spaceN = (n<2) ? 0 : (yn2-yn1)/(n-1);
  }

  // set min and max coord.
  newMinMaxCoord(min(xm,xn),min(min(ym1,ym2),min(yn1,yn2)));
  newMinMaxCoord(max(xm,xn),max(max(ym1,ym2),max(yn1,yn2)));

  // all nodes of first part
  for (unsigned i=1; i!=m+1; i++)
  {
    // new node
    firstN.push_back(*(node=newNode()));
    subgraph->add(*node);

    // coordinates
    if (coord)
      node->coord(Coordinates(xm,ym1+spaceM*(i-1)));
  }

  // all nodes of second part
  for (unsigned i=1; i!=n+1; i++)
  {
    // new node
    secondN.push_back(*(node=newNode()));
    subgraph->add(*node);

    // coordinates
    if (coord)
      node->coord(Coordinates(xn,yn1+spaceN*(i-1)));
  }

  // no edges
  if (m == 0 || n == 0)
    return subgraph;

  constNodeIter it1  = firstN.begin();
  constNodeIter end1 = firstN.end();
  constNodeIter it2  = secondN.begin();
  constNodeIter end2 = secondN.end();

  // edges between all nodes of first to all nodes of second part
  while (it1 != end1)
  {
    while (it2 != end2)
      subgraph->add(*newEdge(*it1, *it2++));

    it1++; it2=secondN.begin();
  }

  return subgraph;
}

SubGraph* RandomGraph::makeComplete(unsigned nodes, int radius, int xM, int yM)
{
  SubGraph* subgraph = new SubGraph(this);

  // no nodes -> nothing to do
  if (nodes == 0)
    return subgraph;

  bool coord=false;
  NodeIter node;

  // set min and max coord.
  newMinMaxCoord(xM-radius,yM-radius);
  newMinMaxCoord(xM+radius,yM+radius);

  if (radius > 0)
    coord = true;

  // all nodes
  for (unsigned i=1; i!=nodes+1; i++)
  {
    subgraph->add(*(node = newNode()));

    // coodinates
    if (coord)
    {
      double x = xM + radius * cos(2*M_PI/nodes*i);
      double y = yM - radius * sin(2*M_PI/nodes*i);
      node->coord(Coordinates(x,y));
    }
  }

  // all possible edges
  NodeIter end = subgraph->nodesEnd();
  NodeIter beforeEnd = (--subgraph->nodesEnd());
  NodeIter j,i=subgraph->nodesBegin();

  while (i != beforeEnd)
  {
    j = i; j++;
    while (j != end)
      subgraph->add(*newEdge(*i,*j++));

    i++;
  }

  return subgraph;
}

SubGraph* RandomGraph::makeHamilton(unsigned nodes, unsigned edges)
{
  if (nodes > edges)
    THROW(runtime_error,"RandomGraph::makeHamilton(unsigned,unsigned)\
      not enough edges for an Hamiltonian cycle")

  else if (nodes*(nodes-1)/2 < edges)
    THROW(runtime_error,"RandomGraph::makeHamilton(unsigned,unsigned)\
      to much edges !")

  SubGraph* subgraph = new SubGraph(this);
  NodeIter first, source, target;
  subgraph->add(*(first = source = newNode()));

  // hamiltonian cycle
  for (unsigned i=2; i!=nodes+1; i++)
  {
    subgraph->add(*(target = newNode()));
    subgraph->add(*newEdge(*source,*target));
    edges--;
    source = target;
  }
  subgraph->add(*newEdge(*first,*target));
  edges--;

  unsigned v1, v2;

  while(edges)
  {
    v1 = (unsigned)((double)nodes*urn())+1;
    v2 = (unsigned)((double)nodes*urn())+1;

    // if the two random nodes are not same and there is no edge between them
    //  -> a new edge is set
    if (v1!=v2 && findEdge(v1,v2)==edgesEnd())
    {
      subgraph->add(*newEdge(v1,v2));
      edges--;
    }
  }

  return subgraph;
}

SubGraph* RandomGraph::makeRandom(unsigned nodes, unsigned edges)
{
  if (nodes*(nodes-1)/2 < edges)
    THROW(runtime_error,"RandomGraph::makeRandom(unsigned,unsigned)\
      to much edges !")

  SubGraph* subgraph = new SubGraph(this);

  for (unsigned i=1; i!=nodes+1; i++)
    subgraph->add(*newNode());

  unsigned v1, v2;

  while(edges)
  {
    v1 = (unsigned)((double)nodes*urn())+1;
    v2 = (unsigned)((double)nodes*urn())+1;

    // if the two random nodes are not same and there is no edge between them
    //  -> a new edge is set
    if (v1!=v2 && findEdge(v1,v2)==edgesEnd())
    {
      subgraph->add(*newEdge(v1,v2));
      edges--;
    }
  }

  return subgraph;
}

SubGraph* RandomGraph::makeErdoesRenyi(unsigned nodes, unsigned edges)
{
  if (nodes*(nodes-1)/2 < edges)
    THROW(runtime_error,"RandomGraph::makeRandom(unsigned,unsigned)\
      to much edges !")

  SubGraph* subgraph = new SubGraph(this);

  for (unsigned i=1; i!=nodes+1; i++)
    subgraph->add(*newNode());

  for (unsigned v1 = 1; v1 < nodes+1; v1++)
    for (unsigned v2 = v1+1; v2 < nodes+1; v2++)
      if ( urn() <= 2*((double)edges/(nodes*(nodes-1))) )
	subgraph->add(*newEdge(v1,v2));

  return subgraph;
}

SubGraph* RandomGraph::makeSmallWorldold(unsigned nodes,  unsigned k, double p,
				      int radius, int xM, int yM)
{
  // <k> = 2*edges/nodes
  // createHeaven();
  // createLand();
  // letThereBeLight();
  if (nodes < 2 || (nodes & 1 && k & 1) || k < 2 || k == nodes-1)
    THROW(runtime_error,"RandomGraph::makeSmallWorld(unsigned,unsigned)\
      can't create initial regular lattice !")

  SubGraph* subgraph = new SubGraph(this);

  bool coord=false;
  NodeIter node;


  // set min and max coord.
  newMinMaxCoord(xM-radius,yM-radius);
  newMinMaxCoord(xM+radius,yM+radius);

  if (radius > 0)
    coord = true;

  // all nodes
  for (unsigned i=1; i!=nodes+1; i++)
  {
    subgraph->add(*(node = newNode()));

    // coodinates
    if (coord)
    {
      double x = xM + radius * cos(2*M_PI/nodes*i);
      double y = yM - radius * sin(2*M_PI/nodes*i);
      node->coord(Coordinates(x,y));
    }
  }
  //srand(2);
  //add edges
  map<Edge,unsigned> rewire;
  for (unsigned j=1; j!=(k/2)+1; j++)
  {
    for (unsigned i=1; i!=nodes-j+1; i++)
      rewire[*newEdge(i,i+j)] = 0;
    for (unsigned i=1; i!=j+1; i++)
      rewire[*newEdge(i,i+nodes-j)] = 0;
  }
  if (k & 1)
    for (unsigned i=1; i!=(nodes/2)+1; i++)	   
      rewire[*newEdge(i,i+(nodes/2))] = 0;

  //  for (constEdgeIter it = edgesBegin();
  //     it != edgesEnd(); it++)
  for (map<Edge,unsigned>::iterator it = rewire.begin();
       it != rewire.end(); it++)
    if (urn() < p)
    {
      unsigned newN;
      do 
      {
	newN = (unsigned)((double)nodes*urn())+1;
      } while (newN == it->first.source().id() ||
	       findEdge(it->first.source().id(),newN) != edgesEnd());
      rewire[it->first] = newN;
      subgraph->add(*newEdge(it->first.source().id(),newN));
    } else {
      subgraph->add(it->first);
    }    

  for (map<Edge,unsigned>::iterator it = rewire.begin();
       it != rewire.end(); it++)
    if (it->second)
    {
      //subgraph->add(*newEdge(it->first.source().id(),it->second));
      delEdge(*findEdge(it->first.source().id(),it->first.target().id()));
    } else {
      //subgraph->add(it->first);
    }

  return subgraph;

}


SubGraph* RandomGraph::makeSmallWorld(unsigned nodes,  unsigned k, double p,
				      int radius, int xM, int yM)
{
  // <k> = 2*edges/nodes
  // createHeaven();
  // createLand();
  // letThereBeLight();
  k *= 2;
  if (nodes < 2 || (nodes & 1 && k & 1) || k < 2 || k == nodes-1)
    THROW(runtime_error,"RandomGraph::makeSmallWorld(unsigned,unsigned)\
      can't create initial regular lattice !")

      //p = 2e/(n(n-1-2k)) wrong
      //p = 2(e-nk)/(n(n-1-2k))

  SubGraph* subgraph = new SubGraph(this);

  bool coord=false;
  NodeIter node;


  // set min and max coord.
  newMinMaxCoord(xM-radius,yM-radius);
  newMinMaxCoord(xM+radius,yM+radius);

  if (radius > 0)
    coord = true;

  // all nodes
  for (unsigned i=1; i!=nodes+1; i++)
  {
    subgraph->add(*(node = newNode()));

    // coodinates
    if (coord)
    {
      double x = xM + radius * cos(2*M_PI/nodes*i);
      double y = yM - radius * sin(2*M_PI/nodes*i);
      node->coord(Coordinates(x,y));
    }
  }
  //srand(2);
  //add edges
  for (unsigned j=1; j!=(k/2)+1; j++)
  {
    for (unsigned i=1; i!=nodes-j+1; i++)
      newEdge(i,i+j);
    for (unsigned i=1; i!=j+1; i++)
      newEdge(i,i+nodes-j);
  }
  if (k & 1)
    for (unsigned i=1; i!=(nodes/2)+1; i++)	   
      newEdge(i,i+(nodes/2));

  for (unsigned v1 = 1; v1 < nodes+1; v1++)
    for (unsigned v2 = v1+1; v2 < nodes+1; v2++)
      if ( urn() <= p )
	subgraph->add(*newEdge(v1,v2));

  return subgraph;

}

SubGraph* RandomGraph::makeScaleFree(double m_, unsigned m0, unsigned t)
{
  // createHeaven();
  // createLand();
  // letThereBeLight();

  //  unsigned m = (unsigned)( ((double)nodes/2)-
  //			   ((double)sqrt((nodes*nodes)-(4*edges))/2) );
  // unsigned m0 = m;

  if (m_ > m0+1)
    THROW(runtime_error,"RandomGraph::makeScaleFree(unsigned,unsigned)\
      m too big !")

  unsigned m = rint(2*m_);
  unsigned add1half = 1;
  if (m%2) 
    m++;
  else
    add1half = 0;
  m /= 2;

  SubGraph* subgraph = new SubGraph(this);
  unsigned nodes= t+m0;
  for (unsigned i=1; i!=nodes+1; i++)
    subgraph->add(*newNode());
  //static int seed = 2;
  //srand(seed++);
  vector<unsigned> degrees;
  set<unsigned> picked;
  degrees.reserve(100);
  for (unsigned i=1; i!=m+1; i++)
  {
    subgraph->add(*newEdge(1,i+1));
    degrees.push_back(1);
    degrees.push_back(i+1);
  }

  for (unsigned i=2+m0; i!=(nodes+1); i++)
  { m -= add1half*(i%2);
    picked.clear();
    for (unsigned j=1; j!=m+1; j++)
    {
      unsigned pick = (unsigned)((double)degrees.size()*urn());
      while (picked.count(degrees[pick]))
      {
	pick++;
	if (pick >= degrees.size())
	  pick = 0;
      }
      subgraph->add(*newEdge(i,degrees[pick]));
      degrees.push_back(i);
      degrees.push_back(degrees[pick]);
      picked.insert(degrees[pick]);
    } m += add1half*(i%2);
  }   

  return subgraph;
}



} // namespace graphlib

/* end of file */
