/***************************************************************************
                          randomgraph.h  -  description
                             -------------------
    begin                : Wed Mar 28 2001
    copyright            : (C) 2001 by Gleiss Petra Manuela
    email                : petra@gleiss.at
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef RANDOMGRAPH_H
#define RANDOMGRAPH_H

#include "graphlib.h"
#include "graph.h"
//#include "subgraph.h"

namespace graphlib {

  // Flag for graphs
  enum GraphFlag { Empty = 0,
                   Bipartite,
                   Complete,
                   Hamiltonian,
                   Random,
                   ErdoesRenyi,
                   SmallWorld,
                   ScaleFree
                 };

class SubGraph //temporary dummy class
{
 public:
  inline SubGraph(const Graph* g) { ; }
  inline SubGraph* add(const Node& n) { ; }
  inline SubGraph* add(const Edge& e) { ; }
  inline NodeIter nodesEnd() { ; }
  inline NodeIter nodesBegin() { ; }
  inline EdgeIter edgesEnd() { ; }
  inline EdgeIter edgesBegin() { ; }
};

#define newMinMaxCoord(x,y) {;}

/**
  *@author Gleiss Petra Manuela
  */

class RandomGraph : virtual public Graph
{
  public:
    //------------------------------------------------------ Con-/Destructors
    /**
      * Constructor. Create a new graph with edges.
      *
      * @param number of nodes
      * @param number of edges
      * @param GraphFlag
      */
    RandomGraph(unsigned nodes=0, unsigned edges=0, 
		double prob=0, GraphFlag flag=Empty);

    /**
      * Destructor. Deletes everything.
      */
    ~RandomGraph();


  private:
    /**
      * CopyConstructor -> not implemented
      */
    RandomGraph(const RandomGraph&);

    /**
      * Assignement operator -> not implemented
      */
    RandomGraph& operator=(const RandomGraph&);

    //protected:
  public:
    //------------------------------------------------------ Generation
    /**
      * Generates a bipartite graph K_(m,n) with <code>m+n</code> nodes.
      * Coordinats will be assigned:
      *  for <code>m</code>: x = xm and y = between ym1 and ym2
      *  for <code>n</code>: x = xn and y = between yn1 and yn2
      * <par>
      * Creates a new subgraph, containing the nodes and edges.
      * <em>Note:</em>: This subgraph is dynamically created with new and
      * will not to be deleted in this function.
      *
      * @return pointer to the subgraph
      */
    SubGraph* makeBipartite(unsigned m, unsigned n,
                            int xm=0, int ym1=0, int ym2=0,
                            int xn=0, int yn1=0, int yn2=0);

    /**
      * Generates a complete graph with <code>nodes</code> nodes.
      * <par>
      * Creates a new subgraph, containing the nodes and edges.
      * <em>Note:</em>: This subgraph is dynamically created with new and
      * will not to be deleted in this function.
      *
      * @return pointer to the subgraph
      */
    SubGraph* makeComplete(unsigned nodes, int radius=0, int xM=0, int yM=0);

    /**
      * Generates a random graph with a hamiltonian cycle.
      * <par>
      * Creates a new subgraph, containing the nodes and edges.
      * <em>Note:</em>: This subgraph is dynamically created with new and
      * will not to be deleted in this function.
      *
      * @return pointer to the subgraph
      */
    SubGraph* makeHamilton(unsigned nodes, unsigned edges);

    /**
      * Generates a random graph.
      * <par>
      * Creates a new subgraph, containing the nodes and edges.
      * <em>Note:</em>: This subgraph is dynamically created with new and
      * will not to be deleted in this function.
      *
      * @return pointer to the subgraph
      */
    SubGraph* makeRandom(unsigned nodes, unsigned edges);

    /**
      * Generates a random Erdoes-Renyi graph.
      * <par>
      * Creates a new subgraph, containing the nodes and edges.
      * <em>Note:</em>: This subgraph is dynamically created with new and
      * will not to be deleted in this function.
      *
      * @return pointer to the subgraph
      */
    SubGraph* makeErdoesRenyi(unsigned nodes, unsigned edges);

    /**
      * Generates a random Watts-Strogatz (small-world) graph.
      * <par>
      * Creates a new subgraph, containing the nodes and edges.
      * <em>Note:</em>: This subgraph is dynamically created with new and
      * will not to be deleted in this function.
      *
      * @return pointer to the subgraph
      */
    SubGraph* makeSmallWorldold(unsigned nodes, unsigned k, double p,
			     int radius=0, int xM=0, int yM=0);

    /**
      * Generates a random Watts-Strogatz (small-world) graph.
      * <par>
      * Creates a new subgraph, containing the nodes and edges.
      * <em>Note:</em>: This subgraph is dynamically created with new and
      * will not to be deleted in this function.
      *
      * @return pointer to the subgraph
      */
    SubGraph* makeSmallWorld(unsigned nodes, unsigned k, double p,
			     int radius=0, int xM=0, int yM=0);

    /**
      * Generates a random Barabasi-Albert (scale-free) graph.
      * <par>
      * Creates a new subgraph, containing the nodes and edges.
      * <em>Note:</em>: This subgraph is dynamically created with new and
      * will not to be deleted in this function.
      *
      * @return pointer to the subgraph
      */

    SubGraph* makeScaleFree(double m, unsigned m0, unsigned t);


}; // class RandomGraph

} // namespace graphlib
#endif // RANDOMGRAPH_H

/* end of file */
