/* $Id: hamiltonmatrix.cpp,v 1.1.1.1 2005/01/04 15:52:03 xtof Exp $ */
/***************************************************************************
                          hamiltonmatrix.cpp  -  description
                             -------------------
    begin                : Wed Oct 14 2001
    copyright            : (C) 2001 by Gleiss Petra Manuela
    email                : petra@gleiss.at
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "hamiltonmatrix.h"
#include "trace.h"
#include "utilities.h"
#include <stdio.h>
#include <errno.h>

#include <algorithm>
#include "eigen.h"
#include "eigen2.h"
#include <set>

using namespace std;

namespace {
/*------------------------------------------------------------------------*/

void nrerror(const char message[])       /* output message upon error */
{
  fprintf(stderr, "\n%s\n", message);
  exit(EXIT_FAILURE);
}

/*-------------------------------------------------------------------------*/
#ifdef DMALLOC
#define space(S) calloc(1,(S))
#else
void *space(unsigned size)
{
  void *pointer;
  
  if ( (pointer = (void *) calloc(1, (size_t) size)) == NULL) {
#ifdef EINVAL
    if (errno==EINVAL) {
      fprintf(stderr,"SPACE: requested size: %d\n", size);
      nrerror("SPACE allocation failure -> EINVAL");
    }
    if (errno==ENOMEM)
#endif
      nrerror("SPACE allocation failure -> no memory");
  }
  return  pointer;
}
#endif
/*------------------------------------------------------------------------*/

char *get_line(FILE *fp) /* reads lines of arbitrary length from fp */
{
  char s[512], *line, *cp;
  
  line = NULL;
  do {
    if (fgets(s, 512, fp)==NULL) break;
    cp = strchr(s, '\n');
    if (cp != NULL) *cp = '\0';
    if (line==NULL)
      line = (char *) space(strlen(s)+1); //line = (char *) calloc(strlen(s)+1,1);  
    else
      line = (char *) realloc(line, strlen(s)+strlen(line)+1);
    strcat(line,s);
   } while(cp==NULL);
  
  return line;
}


list<string> tokenize(const string& S, const string &delim)
{

   list<string> sVector;
   size_t bpos(0);
   while (bpos != std::string::npos)
   {
     size_t epos = S.find_first_of(delim, bpos);
     string token(S.substr(bpos, epos-bpos));
     sVector.push_back(token); //cerr << token;
      if (delim == " ")
         bpos = S.find_first_not_of(delim, epos);
      else
         bpos = (epos == std::string::npos) ? epos : epos + 1;
   }

   return sVector;
}

} 


namespace graphlib {

  //static variables
  map<string,double> HamiltonMatrix::coulomb;
  map< pair<string,string> , double*> HamiltonMatrix::_overlap;
  map< pair<string,string> , double*> HamiltonMatrix::_overlap_alt;
  map<string,double> HamiltonMatrix::Hf;
  double HamiltonMatrix::Ring3;
  double HamiltonMatrix::Ring4;
  double HamiltonMatrix::kappa; 
  double HamiltonMatrix::rep_H_X;
  double HamiltonMatrix::rep_X_X;
  double HamiltonMatrix::V_H;
  double HamiltonMatrix::V_X;

  map< pair<string,string> , char*> HamiltonMatrix::ia;


//---------------------------------------------------------------------------
// Con-/Destructors
//---------------------------------------------------------------------------
HamiltonMatrix::HamiltonMatrix(const Graph& g, 
			       string fname = "", 
			       unsigned n = 0) : 
  LaplacianMatrix(), origGraph(g), smallR(false),
  //origGraph is a reference to g = molecule
  //hamiltonM(laplacianM),                     
  //hamiltonM is reference to laplacianM               
  overlap( (n & 2) ? _overlap_alt : _overlap), hamiltonM(laplacianM), loew(0), eNb(0), nO(n)
{
  sizeM = g.numberNodes();

  if (sizeM == 0)
    THROW(runtime_error,
	  "HamiltonMatrix::HamiltonMatrix(const Graph&) empty graph")
  
  //if params not already given
  if (fname.length()) setParams(fname);

  //expand Graph
  expandGraph(g);

  // initialize a map with nodes and unsigned
  makeNodeNumberMap(graph);

  // initialize hamilton matrix
  makeHamilton(graph);

}

HamiltonMatrix::~HamiltonMatrix()
{ 
  if (overlapM)    delete[] overlapM;
  if (loew) delete [] loew;
}

//---------------------------------------------------------------------------
// Utitlities
//---------------------------------------------------------------------------
void HamiltonMatrix::makeNodeNumberMap(const Graph& g)
{
  nodes.clear();
  
  constNodeIter it  = g.nodesBegin();
  constNodeIter end = g.nodesEnd();
  unsigned int number = 0;

  // iterate through all nodes of the graph
  while (it != end)
  {
    nodes.insert(nodes.end(),make_pair(*it,number++));
    it++;
  }
}

void HamiltonMatrix::makeHamilton(const Graph& g)
{
  if (hamiltonM) delete[] hamiltonM;  
  hamiltonM = new double[sizeM*sizeM];
  fill(hamiltonM,hamiltonM+sizeM*sizeM,0);
  overlapM = new double[sizeM*sizeM];
  fill(overlapM,overlapM+sizeM*sizeM,0);  

  constEdgeIter it  = g.edgesBegin();
  constEdgeIter end = g.edgesEnd();
  unsigned source, target;

  // iterate through all edges of the graph
  while (it != end)
  {
    source = nodes.find(it->source())->second;
    target = nodes.find(it->target())->second;

  if (!overlap.count( pair<string,string>(it->source().label(),
		         		   it->target().label()) ))
    THROW(runtime_error,
          "HamiltonMatrix::makeHamilton(const Graph&) no overlap" + it->source().label() + it->target().label())

    // source -> target
    //cerr << it->source().label() << "-" << it->target().label()  << endl;
    hamiltonM[target + source*sizeM] = hamiltonM[source + target*sizeM] = 
      kappa * (coulomb[it->source().label()] + coulomb[it->target().label()]) *
      overlap[pair<string,string>(it->source().label(),it->target().label())][it->label()[0] - '0'] / 2;
    // target -> source
    //hamiltonM[target + source*sizeM] = 1;
    overlapM[target + source*sizeM] = overlapM[source + target*sizeM] = 
      overlap[pair<string,string>(it->source().label(),it->target().label())][it->label()[0] - '0'];
    it++;
  }
  
  {
  // scale hamiltonian and overlap in 3- and 4-membered rings
  EdgeIter it = Ring3List.begin();
  EdgeIter end = Ring3List.end();
  while (it != end)
  { smallR = true;
    source = nodes.find(it->source())->second;
    target = nodes.find(it->target())->second;

    hamiltonM[target + source*sizeM] *= Ring3;
    hamiltonM[source + target*sizeM] *= Ring3;

    overlapM[target + source*sizeM] *= Ring3;
    overlapM[source + target*sizeM] *= Ring3;
    // change type to 1
    it->label("1");
    it++;
  }
  it = Ring4List.begin();
  end = Ring4List.end();
  while (it != end)
  { smallR = true;
    source = nodes.find(it->source())->second;
    target = nodes.find(it->target())->second;

    hamiltonM[target + source*sizeM] *= Ring4;
    hamiltonM[source + target*sizeM] *= Ring4;

    overlapM[target + source*sizeM] *= Ring4;
    overlapM[source + target*sizeM] *= Ring4;
    // change type to 1
    it->label("1");
    it++;
  }
  it = shareH.begin();
  end = shareH.end();
  while (it != end)
  { smallR = true;
    source = nodes.find(it->source())->second;
    target = nodes.find(it->target())->second;

    hamiltonM[target + source*sizeM] *= Ring3;
    hamiltonM[source + target*sizeM] *= Ring3;

    overlapM[target + source*sizeM] *= Ring3;
    overlapM[source + target*sizeM] *= Ring3;
    // change type to 1
    it->label("1");
    it++;
  }
  }
  
  {
    constNodeIter it = g.nodesBegin();
    constNodeIter end = g.nodesEnd();
    unsigned i;
    // set diagonal items to Hii and 1 respectively
    // in hamilton and overlap matrices
    while (it !=end)
    {
      i = nodes.find(*it)->second;
      hamiltonM[i + (sizeM * i)] = coulomb[it++->label()];
      overlapM[i + (sizeM * i)] = 1.0;
    }
  }

  // if we don't neglect the overlap (S is not I)
  // we'll solve the generalized eigenvalue problem HC=Sce
  // using Loewdin orthogonalization O(N^3)
  // alternative procedure would be Cholesky decomposition O(N)
  // (canonical orthogonalization)
  // J. Chem. Phys., Vol. 110, No. 5, 1 February 1999 Matt Challacombe
  // H is orthogonalized (H -> L^(-1)HL^(-T))
  if (!(nO & 1)) 
  {  
    loew = new double[sizeM*sizeM];
    fill(loew,loew+sizeM*sizeM,0);
    double* tmp = new double[sizeM*sizeM];
    fill(tmp,tmp+sizeM*sizeM,0);
    minvsqrt(loew, overlapM, sizeM);//choldc(chol, overlapM, sizeM); //overlapM must be positive definite
    mmul(tmp, loew, hamiltonM, sizeM);
    trnm(loew, sizeM);
    mmul(hamiltonM, tmp, loew, sizeM); symm(hamiltonM, sizeM);
    if (tmp) delete [] tmp;
  }


}

void HamiltonMatrix::expandGraph(const Graph& g) 
{

  NodeIter newN1, newN2;
  constNodeIter itN, endN;
  constEdgeIter  it, end;

  // temporary map, inverse of origin
  // atom <-> list<orbitals>(H- bonding first)
  //map<Node, list<Node> > derived;

  set<Node>::iterator itN_, endN_;	
  
  set<Node> sp2,sp;
  itN = g.nodesBegin();
  endN = g.nodesEnd();
  while (itN != endN) 
  {
    string h = getHyb(*itN); 
    hyb[itN->id()] = h;
    if (h == "sp2") 
      sp2.insert(*itN);
    if (h == "sp1")  
      sp.insert(*itN);
    itN++;
  }

  // neigbours (with lone pairs) of sp2 hybrids 
  // shall be at least sp2 hybridized
  itN_ = sp2.begin();
  endN_ = sp2.end();
  while (itN_ != endN_)
    { 
    for (constEdgeIter itE = itN_->adjBegin(); itE != itN_->adjEnd(); itE++)
    {
      if (itE->label() == "=")
	continue;
      const Node& neighbor = itN_->opposite(*itE);
      if ( hyb[neighbor.id()] == "sp3" && neighbor.degree() < 4 )
	hyb[neighbor.id()] = "sp2";
    }
    itN_++;
  }
  // neighbours of sp ...
  itN_ = sp.begin();
  endN_ = sp.end();
  while (itN_ != endN_)
  {
    for (constEdgeIter itE = itN_->adjBegin(); itE != itN_->adjEnd(); itE++)
    {
      if (itE->label() == "#")
	continue;
      const Node& neighbor = itN_->opposite(*itE);
      if ( neighbor.label()[0] != 'H' 
	   && (*hyb[neighbor.id()].rbegin()-'0')-(int)neighbor.degree() > 0 )
	hyb[neighbor.id()] = "sp1";
      if ( hyb[neighbor.id()] == "sp3" && neighbor.degree() < 4 )
	hyb[neighbor.id()] = "sp2";
      if ( hyb[neighbor.id()] == "sp2" && neighbor.degree() < 3 )
	hyb[neighbor.id()] = "sp1";
    }
    itN_++;
  }

{
  // detect cycles 
  list<Edge> tmp3;// = new list<Edge>;
  list<Edge> tmp4;// = new list<Edge>;
  {
  set<Edge, less<Edge> > tmpCycEdges(g.edgesBegin(), g.edgesEnd());// = new set<Edge, less<Edge> >(g.edgesBegin(), g.edgesEnd());
  set<Edge, less<Edge> >::iterator it  = tmpCycEdges.begin();
  set<Edge, less<Edge> >::iterator end = tmpCycEdges.end();
  while (it != end)
  {
    constEdgeIter itE1 = it->source().adjBegin();
    constEdgeIter endE1 = it->source().adjEnd();
    bool in3Ring = 0;
    bool in4Ring = 0;
    while (itE1 != endE1)
    {
      const Node& n1 = it->source().opposite(*itE1);
      if (n1 == it->target())
      {
	itE1++;
	continue;
      }
      constEdgeIter itE2 = it->target().adjBegin();
      constEdgeIter endE2 = it->target().adjEnd();
      while (itE2 != endE2)
      {      
	const Node& n2 = it->target().opposite(*itE2);	
	if (n2 == it->source())
	{
	  itE2++;
	  continue;
	}
	if (n1 == n2) {
	  tmp3.push_back(*it);
	  tmp3.push_back(*itE1);
	  tmp3.push_back(*itE2);
	  tmpCycEdges.erase(*itE1);
	  tmpCycEdges.erase(*itE2); 
	  in3Ring = 1;
	  //} else if (g.findEdge(n1.id(),n2.id()) != g.edgesEnd()) 
	  //} else if (n1.findEdge(n2) != n1.adjEnd())
	} else {
	  for (constEdgeIter n1E = n1.adjBegin(); n1E != n1.adjEnd();n1E++)
	  {
	    if (n1.opposite(*n1E) == n2)
	    {
	      in4Ring = 1;
	      break;
	    }
	  }
	}
	itE2++;
      }
      itE1++;
    }
    if (in4Ring && !in3Ring)
      tmp4.push_back(*it); 
    it++;
  }
  //delete tmpCycEdges;
  }

	  
  

  // expand bond -> 2 connected nodes (sigma overlapping orbitals)
  it  = g.edgesBegin();
  end = g.edgesEnd();
  while (it != end)
  {     
      const Node& source = it->source(), target = it->target(); 
      if (source.label()[0] == 'H' && derived[source].size())
      {
	newN1 = graph.findNode(derived[source].begin()->id());
      } else {
	newN1 = addOrbital(source, hyb[source.id()]);
	(source.label()[0] == 'H') ? derived[source].push_front(*newN1) : derived[source].push_back(*newN1);
      }
      if (target.label()[0] == 'H' && derived[target].size())
      {
	newN2 = graph.findNode(derived[target].begin()->id());
      } else {
	newN2 = addOrbital(target, hyb[target.id()]); 
	(target.label()[0] == 'H') ? derived[target].push_front(*newN2) : derived[target].push_back(*newN2);
      }

      // sigma overlapping orbitals = type 0
      EdgeIter newE = graph.newEdge(*newN1,*newN2, "0"); 
      // store sigma orbitals corresponding to bonds in 3/4Rings in lists
      if (count(tmp3.begin(), tmp3.end(),*it))
	Ring3List.push_back(*newE);
      if (count(tmp4.begin(), tmp4.end(),*it))
	Ring4List.push_back(*newE);
							 
      it++;
  }
  //delete tmp3;
  //delete tmp4;
}

  // add lone pairs
  itN = g.nodesBegin();
  endN = g.nodesEnd();
  for (;itN != endN;itN++) 
  {
    if (*itN->label().begin() == 'H')
      if (itN->degree() == 2) {
        constNodeIter H = derived[*itN].begin();
        constEdgeIter H1, H2;
        for (H1 = H->adjBegin(); H1->label()[0] != '0'; H1++) {;}
        H2 = H1;
        for (H2++; H2->label()[0] != '0'; H2++) {;}
        EdgeIter newE = graph.newEdge(H->opposite(*H1),
                                      H->opposite(*H2), "0");
        shareH.push_back(*newE);
        continue;
      } else {
	continue;
      }

    for (int i = (*hyb[itN->id()].rbegin()-'0')+1-(int)itN->degree() ; i ; i--)
    { 
      newN1 = addOrbital(*itN, hyb[itN->id()]); 
      derived[*itN].push_front(*newN1);
      derived_of_type[3][itN->id()].push_back(*newN1);
    }
  }

  // temporary list of edges in expanded graph
  it  = graph.edgesBegin();
  end = graph.edgesEnd(); 
  list<Edge> tmpEdges(it,end);

  // add to edge a->b edges neighbours(a)->b and vice versa (indirect overlapping)
  // (separated from type 0 edge creation in order to have
  // all necessary orbitals established from the start)
  it  = tmpEdges.begin();
  end = tmpEdges.end(); 
  while (it != end)
  { 
    const Node& source = it->source(), target = it++->target(); 

    {
    const list<Node>& tmpNodes = derived[origin[source]];
    itN = tmpNodes.begin();
    endN = tmpNodes.end();
    while (itN != endN)
    {
      if (*itN != source)
	// indirect overlapping = type 2
	graph.newEdge(itN->id(), target.id(), "2");
      itN++;
    }
    }

    {
    const list<Node>& tmpNodes = derived[origin[target]];
    itN = tmpNodes.begin();
    endN = tmpNodes.end();
    while (itN != endN)
    {
      if (*itN != target)
	graph.newEdge(itN->id(), source.id(), "2");
      itN++;
    }   
    }
 
  }

  // add one p orbital to sp2 atoms
  // and add type 0 banana edges to atoms sharing an H
  itN = g.nodesBegin();
  endN = g.nodesEnd();
  while (itN != endN) 
  {   
    if (hyb[itN->id()] == "sp2")
    {
      newN1 = addOrbital(*itN, "p"); 
      derived[*itN].push_back(*newN1);
      if (!sp2.count(*itN)) // was upgraded?
        derived_of_type[3][itN->id()].push_back(*newN1); 
    } /*else if (itN->label()[0] == 'H' && itN->degree() == 2) {
      constNodeIter H = derived[*itN].begin();  
      constEdgeIter H1, H2; 
      for (H1 = H->adjBegin(); H1->label()[0] != '0'; H1++) {;} 
      H2 = H1; 
      for (H2++; H2->label()[0] != '0'; H2++) {;} 
      EdgeIter newE = graph.newEdge(H->opposite(*H1),
				    H->opposite(*H2), "0"); 
      shareH.push_back(*newE);
    } */
    itN++;
  }

  // add nodes (pi orbitals) to sp atoms
  // and add 1 edge between each sp3 and sp2/sp atom pair
  // ( hyperconjugation )
  it = g.edgesBegin();
  end = g.edgesEnd();
  tmpEdges.clear();
  while (it != end)
  {

    const Node& source = it->source(), target = it->target(); 
    char s = *(hyb[source.id()].rbegin()), t = *(hyb[target.id()].rbegin());

    // double bonds (except allenes)
    if (s == '2' && t == '3')
    {
      // random staggered conformation 
      addSymHyperEdges(target, source, derived);
      it++; continue;
    }
    if (t == '2' && s == '3')
    {
      addSymHyperEdges(source, target, derived); 
      it++; continue;
    }
    if (s == '2' && t == '2')
    {
      //pi overlap -> type 1
      graph.newEdge(derived[source].back(), derived[target].back(), "1");
      it++; continue;
    }

    // triple bonds + allenes
    if (s == '1')
    {	
      // random conformation (correct for allene)
      if (t == '2')
      {
	newN1 = addOrbital(source, "p");
	derived[source].push_back(*newN1);
	graph.newEdge(derived[source].back(), derived[target].back(), "1");
	//if (!sp1.count(source)) // was upgraded?
          //derived_of_type[3][itN->id()].push_back(*newN1);
	it++; continue;
      }
      if (t == '3')  
      {    
	if (derived[source].size() < 3)
	{
	  newN1 = addOrbital(source, "p");
	  derived[source].push_back(*newN1);
	}
	addSymHyperEdges(target, source, derived);
	it++; continue;	
      }
    }

    if (t == '1')
    {	
      if (s == '2')
      {
	newN1 = addOrbital(target, "p");
	derived[target].push_back(*newN1);
	graph.newEdge(derived[target].back(), derived[source].back(), "1");
	it++; continue;
      }
      if (s == '3')  
      {    
	if (derived[target].size() < 3)
	{
	  newN1 = addOrbital(target, "p");
	  derived[target].push_back(*newN1);
	}
	addSymHyperEdges(source, target, derived);	
	it++; continue;	
      }
    }

    if (s == '1' && t == '1')
      tmpEdges.push_back(*it);

    it++; 

  }

  // triple bonds, continued
  it = tmpEdges.begin();
  end = tmpEdges.end();
  while (it != end)
  {

    const Node& source = it->source(), target = it++->target(); 
    /* if (derived[source].size() < 4 && derived[target].size() < 4)
    {
      if (derived[source].size() == 2)
      {
	newN1 = addOrbital(source, "p");
	derived[source].push_back(*newN1);
      }
      if (derived[target].size() == 2)
      {
	newN2 = addOrbital(target, "p");
	derived[target].push_back(*newN2);
      }
      //pi overlap -> type 1
      graph.newEdge(derived[source].back(), derived[target].back(), "1");
      newN1 = addOrbital(source, "p");
      derived[source].push_back(*newN1);
      newN2 = addOrbital(target, "p");
      derived[target].push_back(*newN2);
      graph.newEdge(derived[source].back(), derived[target].back(), "1");
      continue;    
      }*/
      if (derived[source].size() == 2)
      {
	newN1 = addOrbital(source, "p");
	derived[source].push_back(*newN1);
      }
      if (derived[target].size() == 2)
      {
	newN2 = addOrbital(target, "p");
	derived[target].push_back(*newN2);
      }
    if (derived[source].size() == 3)
    {
      newN1 = addOrbital(source, "p");
      derived[source].push_back(*newN1);
    }
    if (derived[target].size() == 3)
    {
      newN2 = addOrbital(target, "p");
      derived[target].push_back(*newN2);
    }   
    newN1 = derived[source].end();
    newN2 = derived[target].end();
    graph.newEdge(*--newN1, *--newN2, "1");
    graph.newEdge(*--newN1, *--newN2, "1");
  }
 
  // number of nodes in expanded graph is number of AO
  sizeM = graph.numberNodes();


}

void HamiltonMatrix::setParams(const string& fname)
{
  FILE* fp = 0;
  char *line = 0;
  int pars;

  // open file
  fp=fopen(fname.c_str(),"r");

  // check if really opened
  if (fp == 0)
    throw logic_error("No file opened!");

  // jump to next '/'
  for(int i = 0 ; i < 200 ; i++) {
   line = get_line(fp); 
   //if (!line) { 
    if (line[0] == '/') {
      free(line);
      break;
    }
    free(line);
  }

  // read Atoms
  line = get_line(fp);
  list<string> Atoms = tokenize(line, " ");
  free(line);
  

  for(int i = 0 ; i < 200 ; i++) {
   line = get_line(fp);   
    if (line[0] == '/') {
      free(line);
      break;
    }
    free(line);
  }

  // read coulomb integral parameters
  line = get_line(fp);   
  sscanf(line, "%%%d", &pars);
  parser(fp, line, pars);
  //free(line);


  for(int i = 0 ; i < 200 ; i++) {
   line = get_line(fp); 
   //if (!line) { 
    if (line[0] == '/') {
      free(line);
      break;
    }
    free(line);
  }

  // read sigma overlap matrix parameters
  while (line = 0, line = get_line(fp), line[0] == '%')
  {  
    sscanf(line, "%%%d", &pars); //cerr << Ttos(pars);
    parser(fp, line, pars);
    //free(line);
  }
  //if (line) free(line);
    

  for(int i = 0 ; i < 200 ; i++) {
   line = get_line(fp);   
    if (line[0] == '/') {
      free(line);
      break;
    }
    free(line);
  }

  // read pi overlap matrix parameters
  while (line = 0, line = get_line(fp), line[0] == '%')
  {  
    sscanf(line, "%%%d", &pars); //cerr << Ttos(pars);
    parser(fp, line, pars);
    //free(line);
  }
  //if (line) free(line);

  for(int i = 0 ; i < 200 ; i++) {
   line = get_line(fp);   
    if (line[0] == '/') {
      free(line);
      break;
    } 
    free(line);
  }
  
  // read constants for calculation of remaining integrals
  double ratioX, ratioPS, ratioH, ratiohyp, ratiosym, ratiosp2, ratiosp;
  line = get_line(fp);   
  sscanf(line, " %lf %lf %lf %lf %lf %lf %lf %lf %lf", &ratioX, &ratioPS, &ratioH, &ratiohyp, &ratiosym, 
	                                             &ratiosp2, &ratiosp, &Ring3, &Ring4);
  free(line);

  for(int i = 0 ; i < 200 ; i++) {
   line = get_line(fp);   
    if (line[0] == '/') {
      free(line);
      break;
    } 
    free(line);
  }
  
  // read additional constans
  line = get_line(fp);   
  sscanf(line, " %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf", &scale, &Hf["H"], &Hf["C"], &Hf["N"], &Hf["O"], &Hf["P"], &Hf["S"],
	                                                   &kappa, &rep_H_X, &rep_X_X, &V_H, &V_X);
  free(line);

   
  for(int i = 0 ; i < 200 ; i++) {
   line = get_line(fp);   
    if (line[0] == '/') {
      free(line);
      break;
    } 
    free(line);
  }
  
  // read allowed interactions
  while (line = 0, line = get_line(fp), line[0] == '%')
  {  
    sscanf(line, "%%%d", &pars); 
    parser(fp, line, pars);
    //free(line);
  }
  //if (line) free(line);
  
  fclose(fp);

  // calculate not given overlaps with sp2/sp orbitals from the
  // corresponding overlap with the sp3 orbital
  // ( = sp3 * ratio)
  map< pair<string,string> , double* >::iterator it = overlap.begin();
  map< pair<string,string> , double* >::iterator end = overlap.end();
  list< pair<const pair<string,string> , double*> > tmpMap;
  while (it != end)
  {
    if (*(it->first.first.rbegin()) == '3') 
    {
      pair<string,string> ov_pair(*(it->first.first.begin()) + string("sp2"),it->first.second);
      if (!overlap.count(ov_pair))
      {
	overlap[ov_pair] = new double[3];
	overlap[ov_pair][0] = it->second[0] * ratiosp2;
      }
      ov_pair = pair<string,string>(*(it->first.first.begin()) + string("sp1"),it->first.second);
      if (!overlap.count(ov_pair))
      {
	overlap[ov_pair] = new double[3];
	overlap[ov_pair][0] = it->second[0] * ratiosp;
      }	
      //      it++;
      //      continue;
    }
    if (*(it->first.second.rbegin()) == '3') 
    {
      pair<string,string> ov_pair(it->first.first,*(it->first.second.begin()) + string("sp2"));
      if (!overlap.count(ov_pair))
      {
	overlap[ov_pair] = new double[3];
	overlap[ov_pair][0] = it->second[0] * ratiosp2;
      }
      ov_pair = pair<string,string>(it->first.first,*(it->first.second.begin()) + string("sp1"));
      if (!overlap.count(ov_pair))
      {
	overlap[ov_pair] = new double[3];
	overlap[ov_pair][0] = it->second[0] * ratiosp;
      }	
      //      it++;
      //      continue;
    }
    it++;
  }


  // calculate indirect overlaps ( = direct * ratio)
  it = overlap.begin();
  end = overlap.end();
  while (it != end) 
  {
    if (*it->first.first.rbegin() == 'p' && *it->first.second.rbegin() == 'p')
    {
      // store p/p-orbital overlaps for later
      tmpMap.push_back(*it);
      it++;
    } else if (*it->first.first.rbegin() == 'p' || *it->first.second.rbegin() == 'p')
    {
      // overlaps with p-orbitals will be calculated lateron
      it++;
    } else if (it->first.first == "Hs" || it->first.second == "Hs")
    {
      it->second[2] = it++->second[0]*ratioH;
    } else if (*it->first.first.begin() == 'P' || *it->first.second.begin() == 'P' ||
               *it->first.first.begin() == 'S' || *it->first.second.begin() == 'S' )
    {
      it->second[2] = it++->second[0]*ratioPS;
    } else {
      it->second[2] = it++->second[0]*ratioX; 
    }
  }

  // calculate overlaps with p-orbitals using p/p-orbital overlaps
  // hyperconjugation: x/p = p/p * ratiohyp
  // symmetric: x/p = p/p * ratiosym
  // the latter is zero theoretically but can replace the hyperconjugation
  // in case we want the energy to be reproducible (not conformation-based)
  list< pair<const pair<string,string> , double*> >::iterator itM = tmpMap.begin();
  list< pair<const pair<string,string> , double*> >::iterator endM = tmpMap.end();
  while (itM != endM)
  {

      double PS = (*itM->first.second.begin() == 'P' || *itM->first.second.begin() == 'S')? ratioPS/ratioX : 1.;
      string second = itM->first.second.substr(0,1) + "sp3";

	if (!overlap.count(pair<string,string>(itM->first.first, second)))
	  overlap[pair<string,string>(itM->first.first, second)] = new double[3];
	if (!overlap.count(pair<string,string>(second, itM->first.first)))
	  overlap[pair<string,string>(second, itM->first.first)] = new double[3];
	overlap[pair<string,string>(itM->first.first, second)][1] = itM->second[1]*ratiohyp*PS;
	overlap[pair<string,string>(itM->first.first, second)][2] = itM->second[1]*ratiosym*PS;
	overlap[pair<string,string>(second, itM->first.first)][1] = itM->second[1]*ratiohyp*PS;
	overlap[pair<string,string>(second, itM->first.first)][2] = itM->second[1]*ratiosym*PS;
	//cerr << itM->first.first <<  second;
      itM++;
  }
  
  //overlap[pair<string,string>(string("Hs"),string("Hs"))][1] = 0.0;
}

void HamiltonMatrix::printParams()
{
  map< pair<string,string> , double* >::iterator it = overlap.begin();
  map< pair<string,string> , double* >::iterator end = overlap.end();
  while (it != end) 
  {
    cerr << it->first.first << "-" << it->first.second << " " << Ttos(it->second[0]) << " " << Ttos(it->second[1]) << " " << Ttos(it->second[2]) << endl;
    //if (it->second[0] > 0.85 || it->second[1] > 0.85 || it->second[2] > 0.85)
    //cerr << "Overlap over 0.85 for " << it->first.first << "-" << it->first.second << 
    //" " << Ttos(it->second[0]) << " " << Ttos(it->second[1]) << " " << Ttos(it->second[2]) << endl;
    it++;
  }
}

void HamiltonMatrix::freeParams()
{
  coulomb.clear();
  map< const pair<string,string> , double* >::iterator itM = overlap.begin();
  map< const pair<string,string> , double* >::iterator endM = overlap.end();
  while (itM != endM) 
  {
    free(itM++->second);
  }
  overlap.clear();
}

inline string HamiltonMatrix::getHyb(const Node& atom)
{
  int lone_eNb = 0;
  bool H = false;
  string::iterator label_it = atom.label().begin();
  switch ( *label_it ) 
  {
    case 'H' :
      eNb += 1;
      lone_eNb = 2;
      H = true;
      break;
      
    case 'C' :
      eNb += 4;
      lone_eNb = 8;
      break;

    case 'N' :
      eNb += 5;      
      lone_eNb = 10;
      break; 
     
    case 'O' :
      eNb += 6;      
      lone_eNb = 12;	
      break;

    case 'P' :
      eNb += 5;
      lone_eNb = 10;
      break;

   case 'S' :
      eNb += 6;
      lone_eNb = 12;
      break;
  }

  constEdgeIter itE = atom.adjBegin();
  constEdgeIter endE = atom.adjEnd();
  while (itE != endE)
  {
    if (itE->label() == "")
    {
      lone_eNb -= 2;
      itE++;
      continue;
    }
    switch (*(itE->label().begin()))
    {
      case '-' :
        lone_eNb -= 2;
	break;
      case '=' :
        lone_eNb -= 4; 
	break;
      case ':' :
        lone_eNb -= 3;
	break;
      case '#' :
        lone_eNb -= 6;
	break;
    }
    itE++;
  }

  if (++label_it != atom.label().end())
  {  
    unsigned charge;
    if ( *(label_it) == '+')
    {
      if (++label_it == atom.label().end())
	{ charge = 1; }
      else 
	{ charge = *(label_it) - '0'; }
      eNb -= 2*charge; 
      lone_eNb -= 2*charge;
    } else { if ( *(label_it) == '-') {
      if (++label_it == atom.label().end())
	{ charge = 1; }
      else 
	{ charge = *(label_it) - '0'; }
      eNb += 2*charge;
      lone_eNb += 2*charge;
    } }
  }

  if (H)
    return string("s");

  if (lone_eNb < 0) 
    THROW(runtime_error,
          "HamiltonMatrix::getHyb() negative number of lone e pairs:" + Ttos(lone_eNb) + " on " + atom.label() + " in " + origGraph.name)

  return("sp" + Ttos( atom.degree()-1+(lone_eNb/4) ));
}

inline NodeIter HamiltonMatrix::addOrbital(const Node& atom, string hyb)
{
  NodeIter newN; 
  newN = graph.newNode(); 
  newN->label(atom.label()[0] + hyb);
  origin[*newN] = atom;
  if (hyb == "sp3") 
    derived_of_type[0][atom.id()].push_back(*newN);
  if (hyb == "s") 
    derived_of_type[1][atom.id()].push_back(*newN);
  if (hyb == "p") 
    derived_of_type[2][atom.id()].push_back(*newN);
  return newN;
}

void HamiltonMatrix::addSymHyperEdges(const Node& source, 
				      const Node& target,
				      map< Node, list<Node> >& derived)
{
      constNodeIter it = derived[source].begin();
      constNodeIter end = derived[source].end();
      while (it != end)
      {
	if (*it->label().rbegin() != '3') 
	  continue;
	constEdgeIter itE; 
	for (itE = it->adjBegin(); itE != it->adjEnd(); itE++)
	{
	  if (itE->label() == "0" && find(shareH.begin(),shareH.end(),*itE) == shareH.end()) 
	    break;
	} 
	if (itE != it->adjEnd() && origin[it->opposite(*itE)] == target)
	  break;
	it++;
      }
      if (it == end)
        THROW(runtime_error,
              "HamiltonMatrix::addSymHyperEdges() no sigma bond found between " + Ttos(source.id()) + " and " + Ttos(target.id()) + " in " + origGraph.name)
	
 
      if (*it == derived[source].back())
      {
	graph.newEdge(derived[target].back(), derived[source].back(), "2");
	graph.newEdge(derived[target].back(), *(++derived[source].rbegin()), "1");
      } else {
	graph.newEdge(derived[target].back(), derived[source].back(), "1");
	graph.newEdge(derived[target].back(), *it, "2");
      }	
}


int HamiltonMatrix::parser(FILE* fp, char* line, int pars)
{


  if (pars == 0)
  {
    int row = 0, tmp = 0, column = 0;
    char* tmpline = 0;
    list<string>::iterator itS, endS;
    list<string> columns;
    string rows;

    sscanf(line, "%%%*d %n%*s%n : %n%*s", &row, &tmp, &column);
    if (!column) { cerr << "invalid overlap matrix row format (skipping following line): " << line << endl; free(get_line(fp)); return -1;  }
    columns.clear();
    columns = tokenize(line+column, " ");
    line[tmp] = '\0';
    rows = string(line + row);
    free(line);
    line = get_line(fp);
    tmpline = line;
    itS = columns.begin();
    endS = columns.end();
    while (itS != endS) 
    {
      tmp = 0;
      if (!overlap.count(pair<string,string>(rows, *itS)))
	  overlap[pair<string,string>(rows, *itS)] = new double[3];
      sscanf(line, " %lf%n", overlap[pair<string,string>(rows, *itS)], &tmp);
      //overlap[pair<string,string>(rows, *itS)][1] = 0.05;
      if (!tmp) cerr << "missing parameter overlap " << rows << "-" << *itS << endl; 
      line += tmp; itS++;
    }
    free(tmpline);
    return(0);
  }

  if (pars == 2)
  {
    int row = 0, tmp = 0, column = 0;
    char* tmpline = 0;
    list<string>::iterator itS, endS;
    list<string> columns;
    string rows;

    sscanf(line, "%%%*d %n%*s%n : %n%*s", &row, &tmp, &column);
    if (!column) { cerr << "invalid overlap matrix row format (skipping following line): " << line << endl; free(get_line(fp)); return -1;  }
    columns.clear();
    columns = tokenize(line+column, " ");
    line[tmp] = '\0';
    rows = string(line + row);
    free(line);
    line = get_line(fp);
    tmpline = line;
    itS = columns.begin();
    endS = columns.end();
    while (itS != endS) 
    {
      tmp = 0;
      if (!overlap.count(pair<string,string>(rows, *itS)))
	  overlap[pair<string,string>(rows, *itS)] = new double[3];
      sscanf(line, " %lf%n", overlap[pair<string,string>(rows, *itS)] + 1, &tmp);
      if (!tmp) cerr << "missing parameter overlap " << rows << "-" << *itS << endl; 
      line += tmp; itS++;
    }
    free(tmpline);
    return(0);
  }

  if (pars == 1)
  {
    int tmp = 0;
    char* tmpline = 0;
    list<string>::iterator itS, endS;
    list<string> columns;

    sscanf(line, "%%%*d %n", &tmp);
    //if (!co) { cerr << "invalid overlap matrix row format (skipping following line): " << line << endl; free(get_line(fp)); return -1;  }
    columns.clear();
    columns = tokenize(line+tmp, " ");
    line[tmp] = '\0';
    free(line);
    line = get_line(fp);
    tmpline = line;
    itS = columns.begin();
    endS = columns.end();
    while (itS != endS) 
    {
      tmp = 0;
      sscanf(line, " %lf%n", &coulomb[*itS], &tmp);
      if (!tmp) cerr << "missing parameter coulomb integral " << *itS << endl; 
      line += tmp; itS++;   
    }
    free(tmpline);   
    return(0);
  }
  
  if (pars == 3)
  {
    int row = 0, tmp = 0, column = 0;
    char* tmpline = 0;
    list<string>::iterator itS, endS;
    list<string> columns;
    string rows;

    sscanf(line, "%%%*d %n%*s%n : %n%*s", &row, &tmp, &column);
    if (!column) { cerr << "invalid overlap matrix row format (skipping following line): " << line << endl; free(get_line(fp)); return -1;  }
    columns.clear();
    columns = tokenize(line+column, " ");
    line[tmp] = '\0';
    rows = string(line + row);
    free(line);
    line = get_line(fp);
    tmpline = line;
    itS = columns.begin();
    endS = columns.end();
    for (unsigned i = 0 ; i < strlen(line) ; i++)
    {
      switch( line[i] )
      {
        case 'b' :
	line[i] = '0';
	break;

        case 's' :
	line[i] = '1';
	break;

        case 'p' :
	line[i] = '2';
	break;

        case 'l' :
	line[i] = '3';
	break;
      }
    }

    while (itS != endS) 
    {
      tmp = 0;
      if (!ia.count(pair<string,string>(rows, *itS)))
        ia[pair<string,string>(rows, *itS)] = new char[10];
      sscanf(line, " %s%n", ia[pair<string,string>(rows, *itS)], &tmp);
      //overlap[pair<string,string>(rows, *itS)][1] = 0.05;
      if (!tmp) cerr << "missing parameter overlap " << rows << "-" << *itS << endl; 
      line += tmp; itS++;
    }
    free(tmpline);
    return(0);
  }
  
  return -1;
}

} // namespace graphlib

/* end of file */



