/***************************************************************************
                          adjacencymatrix.h  -  description
                             -------------------
    begin                : Wed Mar 14 2001
    copyright            : (C) 2001 by Gleiss Petra Manuela
    email                : petra@gleiss.at
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef HAMILTONMATRIX_H
#define HAMILTONMATRIX_H

#include "graph.h"
#include "laplacianmatrix.h"
#include "trace.h"
#include <map>

namespace graphlib {

typedef map<Node,unsigned>  NodeNumberMap;


/**
  * @short hamilton matrix and diagoal matrix of vertex degrees.
  */

 class HamiltonMatrix : public LaplacianMatrix
{
  	//------------------------------------------------------ friend classes
  //friend class LaplacianMatrix;

  	//------------------------------------------------------ friend functions
    /**
      * Output-operator.
      *
      * @param os the output-stream
      * @param a AdjacencMatrix
      * @return the output-stream
      */
      friend ostream& operator<<(ostream& os,const LaplacianMatrix& l);

  public:
  	//------------------------------------------------------ Con-/Destructors
    /**
      * Constructor. Creates new hamilton matrix and diagonal matrix
      * of vertex degrees of <code>graph</code>.
      * <em>Please note:</em> <code>graph</code> may not by empty </em>,
      * otherwise a runtime_error will be thrown.
      * Parameters are read from <code>fname</code> unless it is
      * the empty string or omitted. Bit 2 of n decides which
      * overlap parameter set to use (_overlap / _overlap_alt).
      * Normally, use once 
      * HamiltonMatrix(dummy, "parameter file");
      * and everafter use
      * HamiltonMatrix(graph);
      * @param graph
      */
    HamiltonMatrix(const Graph& graph, string fname, unsigned n);

    /**
      * CopyConstructor. Makes a copy of the original HamiltonMatrix.
      * Allocates new memory!
      */
    HamiltonMatrix(const HamiltonMatrix&);

    /**
      * Destructor. Deletes both matrices.
      */
    ~HamiltonMatrix();

  	//------------------------------------------------------ Operators
    /**
      * Assignement operator. Assigns a copy of the original
      * HamiltonMatrix. Allocates new memory!
      */
    HamiltonMatrix& operator=(const HamiltonMatrix&);


  	
  	//------------------------------------------------------ Information
  	/**
  	  * Returns the size of the matrix.
  	  */
	  unsigned size() const { return sizeM; }

  	/**
  	  * Returns the size of the matrix.
  	  */
    const double* hamilton() const { return hamiltonM; }

    /**
      * Returns a map so that each node is associated with an
      * ascending number
      */
    const NodeNumberMap& nodeNumber() const {return nodes; }

  	/**
  	  * Returns number of electrons
  	  */
    const unsigned eNbr() const { return eNb; }

    /**
      * Returns a map so that each node is associated with an
      * ascending number
      */
    const map<Node,Node>& orbAtom() const {return origin; }

        /**
          * Returns number of electrons
          */
    const bool smallRings() const 
    { 
	return (!Ring3List.empty() || !Ring4List.empty()); 
    }

        //------------------------------------------------------ Output
        /**
          * Writes the adjacency and degree matrix to <code>out</code>.
          * Format: mathematic
          * lapl = { { .. , .. } , { .. , .. } };
          *
          * @param ostream (default stdout)
          */
    void writeMathematica(ostream& c) 
    {
      c.precision(20); 
      c.setf(ios::showpos); 
      LaplacianMatrix::writeMathematica(c);
    }


  	//------------------------------------------------------ datas
	//original Graph, expanded Graph
	const Graph& origGraph;
	Graph graph;
	map < Node, list<Node> > derived;
	// map Node <-> hybridization
	map<int, string> hyb;
	map< int , list<Node> > derived_of_type[4];
	static map< pair<string,string> , char*> ia;

	bool smallR;
	/// parameter maps of label <-> coulomb and 
	/// overlap integrals (sigma, p-pi/hyper,spn-indirect/sym)  , static 
	static map<string,double> coulomb;
	static map< pair<string,string> , double*> _overlap;
	static map< pair<string,string> , double*> _overlap_alt;
	map< pair<string,string> , double*>& overlap;
	static double kappa;

  protected:

  	/// hamilton, overlap, cholesky* matrix
        double*& hamiltonM;
	double* overlapM;
	double* loew;
	/// collateral information: number of electrons
	unsigned eNb;
	/// map original Node <-> orbital-node
	map<Node,Node> origin;
	/// bit 0 : neglect overlap -> HC=Ce, quantum effect scale factor
	/// bit 1 : use _overlap or _overlap_alt parameters
	unsigned nO;
	/// energy constants, static
	static map<string,double> Hf;
	double scale;
	static double Ring3, Ring4, rep_H_X, rep_X_X, V_H, V_X;
        /// edges in 3-, 4-membered rings
        list<Edge> Ring3List, Ring4List, shareH;


  	//------------------------------------------------------ Utilities
    /**
      * Makes a copy of <code>adjaMat</code>.
      *
      * @param adjaMat
      */
    void copy(const HamiltonMatrix& adjaMat);

    /**
      * Creates a map so that each node is associated with an
      * ascending number
      *
      * @param graph
      */
    void makeNodeNumberMap(const Graph&);

    /**
      * Creates the hamilton matrix
      *
      * @param graph
      */
    void makeHamilton(const Graph&);

    /**
     * expands each non-H node to set of four
     * orbital-nodes
     */
    void expandGraph(const Graph&);

    /**
     * reads overlap and coulomb integral
     * parameters in
     */
    void setParams(const string&);

 public:
    /**
     * free memory of overlap and coulomb integral
     * parameters in
     */
    void freeParams();

    /**
     * print Parameters
     */
    void printParams();

    const Node& origAtom( const Node& orb) { return origin[orb]; }

    //  pair< map<Node,Node>::iterator, map<Node,Node>::iterator > 
    //  derived(const Node& atom) { return origin.equal_range(atom); }

 protected:
    /**
     * parse parameters
     */
    int parser(FILE*, char*, int);

    /**
     * get VSEPR hybridization
     */
    string getHyb(const Node&);

    /**
     * add orbital with 'hyb' hybridization
     */
    NodeIter addOrbital(const Node& atom, string hyb);

    /**
     * add type sym and hyper overlap/edges between 
     * p orbital and orbitals of sp3 atom
     */
    void addSymHyperEdges(const Node& source, 
			  const Node& target,
			  map< Node, list<Node> >& derived);

}; // class HamiltonMatrix

} // namespace graphlib

#endif // HAMILTONMATRIX_H

/* end of file */
