/*
  Last changed Time-stamp: <2003-06-12 20:11:09 xtof>
  $Id: vecmath.h,v 1.1.1.1 2005/01/04 15:52:03 xtof Exp $
*/

#ifndef _VECMATH_H_
#define _VECMATH_H_

#include <iostream>
#include <cmath>

using namespace std;

class Point3;

class Vector3 {
 public:
  Vector3() : x(0), y(0), z(0) {}
  Vector3(const Vector3& v) : x(v.x), y(v.y), z(v.z) {}
  Vector3(double _x, double _y, double _z) : x(_x), y(_y), z(_z) {}

  Vector3& operator=(const Vector3& a) {
    x = a.x; y = a.y; z = a.z;
    return *this;
  }

  double operator[](int n) const { return ((double *) this)[n]; }

  Vector3& operator+=(const Vector3& a) {
    x += a.x; y += a.y; z += a.z;
    return *this;
  }

  Vector3& operator-=(const Vector3& a) {
    x -= a.x; y -= a.y; z -= a.z;
    return *this;
  }

  Vector3& operator*=(double s) {
    x *= s; y *= s; z *= s;
    return *this;
  }

  Vector3 operator-() const {
    return Vector3(-x, -y, -z);
  }

  Vector3 operator+() const {
    return *this;
  }

  double length() const {
    return (double) sqrt(x * x + y * y + z * z);
  }

  double lengthSquared() const {
    return x * x + y * y + z * z;
  }

  void normalize() {
    double s = 1.0 / (double) sqrt(x * x + y * y + z * z);
    x *= s; y *= s; z *= s;
  }

  double x, y, z;
};

class Point3 {
 public:
  Point3() : x(0), y(0), z(0) {}
  Point3(const Point3& p) : x(p.x), y(p.y), z(p.z) {}
  Point3(double _x, double _y, double _z) : x(_x), y(_y), z(_z) {}

  Point3& operator=(const Point3& a) {
    x = a.x; y = a.y; z = a.z;
    return *this;
  }

  double operator[](int n) const { return ((double *) this)[n]; }

  Point3& operator+=(const Vector3& v) {
    x += v.x; y += v.y; z += v.z;
    return *this;
  }

  Point3& operator-=(const Vector3& v) {
    x -= v.x; y -= v.y; z -= v.z;
    return *this;
  }

  Point3& operator*=(double s) {
    x *= s; y *= s; z *= s;
    return *this;
  }

  double distanceTo(const Point3& p) const {
    return (double) sqrt((p.x - x) * (p.x - x) +
                         (p.y - y) * (p.y - y) +
                         (p.z - z) * (p.z - z));
  }

  double distanceToSquared(const Point3& p) const {
    return ((p.x - x) * (p.x - x) +
            (p.y - y) * (p.y - y) +
            (p.z - z) * (p.z - z));
  }

  double distanceFromOrigin() const {
    return (double) sqrt(x * x + y * y + z * z);
  }

  double distanceFromOriginSquared() const {
    return x * x + y * y + z * z;
  }

  double x, y, z;
};

class Matrix3 {
 public:
  Matrix3() {
    r[0] = Vector3(0, 0, 0);
    r[1] = Vector3(0, 0, 0);
    r[2] = Vector3(0, 0, 0);
  }

  Matrix3(const Vector3& r0, const Vector3& r1, const Vector3& r2) {
    r[0] = r0;
    r[1] = r1;
    r[2] = r2;
  }

  Matrix3(const Matrix3& m) {
    r[0] = m.r[0];
    r[1] = m.r[1];
    r[2] = m.r[2];
  }

  Matrix3& operator=(const Matrix3& m) {
    r[0] = m.r[0];
    r[1] = m.r[1];
    r[2] = m.r[2];
    return *this;
  }

  const Vector3& operator[](int n) const {
    return r[n];
  }

  Vector3 row(int n) const {
    return r[n];
  }

  Vector3 column(int n) const {
    return Vector3(r[0][n], r[1][n], r[2][n]);
  }

  Matrix3& operator*=(double s) {
    r[0] *= s;
    r[1] *= s;
    r[2] *= s;
    return *this;
  }

  static Matrix3 identity() {
    return Matrix3(Vector3(1, 0, 0),
                   Vector3(0, 1, 0),
                   Vector3(0, 0, 1));
  }

  static Matrix3 rotationXYZtoUVW(Vector3 u, Vector3 v, Vector3 w) {
    return Matrix3(Vector3(u.x, v.x, w.x),
                   Vector3(u.y, v.y, w.y),
                   Vector3(u.z, v.z, w.z));
  }

  Matrix3 transpose() const {
    return Matrix3(Vector3(r[0].x, r[1].x, r[2].x),
                   Vector3(r[0].y, r[1].y, r[2].y),
                   Vector3(r[0].z, r[1].z, r[2].z));
  }

  static double det2x2(double a, double b, double c, double d) {
    return a * d - b * c;
  }

  double determinant() const {
    return (r[0].x * r[1].y * r[2].z +
	    r[0].y * r[1].z * r[2].x +
	    r[0].z * r[1].x * r[2].y -
	    r[0].z * r[1].y * r[2].x -
	    r[0].x * r[1].z * r[2].y -
	    r[0].y * r[1].x * r[2].z);
  }

  Matrix3 inverse() const {
    Matrix3 adjoint;

    adjoint.r[0].x =  det2x2(r[1].y, r[1].z, r[2].y, r[2].z);
    adjoint.r[0].y = -det2x2(r[1].x, r[1].z, r[2].x, r[2].z);
    adjoint.r[0].z =  det2x2(r[1].x, r[1].y, r[2].x, r[2].y);
    adjoint.r[1].x = -det2x2(r[0].y, r[0].z, r[2].y, r[2].z);
    adjoint.r[1].y =  det2x2(r[0].x, r[0].z, r[2].x, r[2].z);
    adjoint.r[1].z = -det2x2(r[0].x, r[0].y, r[2].x, r[2].y);
    adjoint.r[2].x =  det2x2(r[0].y, r[0].z, r[1].y, r[1].z);
    adjoint.r[2].y = -det2x2(r[0].x, r[0].z, r[1].x, r[1].z);
    adjoint.r[2].z =  det2x2(r[0].x, r[0].y, r[1].x, r[1].y);
    adjoint *= 1 / determinant();

    return adjoint;
  }

  Vector3 r[3];
};

class Vector4 {
 public:
  Vector4() : x(0), y(0), z(0), w(0) {}
  Vector4(const Vector4& v) : x(v.x), y(v.y), z(v.z), w(v.w) {}
  Vector4(double _x, double _y, double _z, double _w) :
    x(_x), y(_y), z(_z), w(_w) {}

  Vector4& operator=(const Vector4& a) {
    x = a.x; y = a.y; z = a.z; w = a.w;
    return *this;
  }

  double operator[](int n) const {
    return ((double*) this)[n];
  }

  Vector4& operator+=(const Vector4& a) {
    x += a.x; y += a.y; z += a.z; w += a.w;
    return *this;
  }

  Vector4& operator-=(const Vector4& a) {
    x -= a.x; y -= a.y; z -= a.z; w -= a.w;
    return *this;
  }

  Vector4& operator*=(double s) {
    x *= s; y *= s; z *= s; w *= s;
    return *this;
  }

  Vector4 operator-() {
    return Vector4(-x, -y, -z, -w);
  }

  Vector4 operator+() {
    return *this;
  }

  double x, y, z, w;
};

class Matrix4 {
 public:

  Matrix4() {
    r[0] = Vector4(0, 0, 0, 0);
    r[1] = Vector4(0, 0, 0, 0);
    r[2] = Vector4(0, 0, 0, 0);
    r[3] = Vector4(0, 0, 0, 0);
  }

  Matrix4(const Vector4& v0, const Vector4& v1,
	  const Vector4& v2, const Vector4& v3) {
    r[0] = v0;
    r[1] = v1;
    r[2] = v2;
    r[3] = v3;
  }

  Matrix4(const Matrix4& m) {
    r[0] = m.r[0];
    r[1] = m.r[1];
    r[2] = m.r[2];
    r[3] = m.r[3];
  }

  Matrix4& operator=(const Matrix4& m) {
    r[0] = m.r[0];
    r[1] = m.r[1];
    r[2] = m.r[2];
    r[3] = m.r[3];
    return *this;
  }

  const Vector4& operator[](int n) const {
    return r[n];
  }

  Vector4 row(int n) const {
    return r[n];
  }

  Vector4 column(int n) const {
    return Vector4(r[0][n], r[1][n], r[2][n], r[3][n]);
  }

  static Matrix4 identity() {
    return Matrix4(Vector4(1, 0, 0, 0),
		   Vector4(0, 1, 0, 0),
		   Vector4(0, 0, 1, 0),
		   Vector4(0, 0, 0, 1));
  }

  static Matrix4 translation(const Point3& p) {
    return Matrix4(Vector4(1, 0, 0, p.x),
		   Vector4(0, 1, 0, p.y),
		   Vector4(0, 0, 1, p.z),
		   Vector4(0, 0, 0, 1));
  }

  static Matrix4 translation(const Vector3& v) {
    return Matrix4(Vector4(1, 0, 0, v.x),
		   Vector4(0, 1, 0, v.y),
		   Vector4(0, 0, 1, v.z),
		   Vector4(0, 0, 0, 1));
  }

  static Matrix4 rotation(const Vector3& u, const Vector3& v,
			  const Vector3& w) {
    return Matrix4(Vector4(u.x, u.y, u.z, 0),
		   Vector4(v.x, v.y, v.z, 0),
		   Vector4(w.x, w.y, w.z, 0),
		   Vector4(0  , 0  , 0  , 1));
  }

  static Matrix4 rotation(const Vector3& axis, double angle) {
    // fill in your code here
    Vector3 a = axis;
    a.normalize();
    double c = cos(angle);
    double s = sin(angle);
    double t = 1 - c;

    return Matrix4(Vector4(t * a.x * a.x + c,
			   t * a.x * a.y - s * a.z,
			   t * a.x * a.z + s * a.y,
			   0),
		   Vector4(t * a.x * a.y + s * a.z,
			   t * a.y * a.y + c,
			   t * a.y * a.z - s * a.x,
			   0),
		   Vector4(t * a.x * a.z - s * a.y,
			   t * a.y * a.z + s * a.x,
			   t * a.z * a.z + c,
			   0),
		   Vector4(0, 0, 0, 1));
  }

  static Matrix4 xrotation(double angle) {
    double c = cos(angle);
    double s = sin(angle);

    return Matrix4(Vector4(1, 0,  0, 0),
		   Vector4(0, c, -s, 0),
		   Vector4(0, s,  c, 0),
		   Vector4(0, 0,  0, 1));
  }

  static Matrix4 yrotation(double angle) {
    double c = cos(angle);
    double s = sin(angle);

    return Matrix4(Vector4( c, 0, s, 0),
		   Vector4( 0, 1, 0, 0),
		   Vector4(-s, 0, c, 0),
		   Vector4( 0, 0, 0, 1));
  }

  static Matrix4 zrotation(double angle) {
    double c = cos(angle);
    double s = sin(angle);

    return Matrix4(Vector4(c, -s, 0, 0),
		   Vector4(s,  c, 0, 0),
		   Vector4(0,  0, 1, 0),
		   Vector4(0,  0, 0, 1));
  }

  static Matrix4 scaling(const Vector3& s) {
    return Matrix4(Vector4(s.x, 0  , 0  , 0),
		   Vector4(0  , s.y, 0  , 0),
		   Vector4(0  , 0  , s.z, 0),
		   Vector4(0  , 0  , 0  , 1));
  }

  static Matrix4 scaling(double scale) {
    return scaling(Vector3(scale, scale, scale));
  }

  Matrix4 transpose() const {
    return Matrix4(Vector4(r[0].x, r[1].x, r[2].x, r[3].x),
		   Vector4(r[0].y, r[1].y, r[2].y, r[3].y),
		   Vector4(r[0].z, r[1].z, r[2].z, r[3].z),
		   Vector4(r[0].w, r[1].w, r[2].w, r[3].w));
  }

  Matrix4 inverse() const;

  Vector4 r[4];
};

//
// **** Vector3 operators ****

inline Vector3 operator+(const Vector3& a, const Vector3& b) {
  return Vector3(a.x + b.x, a.y + b.y, a.z + b.z);
}

inline Vector3 operator-(const Vector3& a, const Vector3& b) {
  return Vector3(a.x - b.x, a.y - b.y, a.z - b.z);
}

inline Vector3 operator*(double s, const Vector3& v) {
  return Vector3(s * v.x, s * v.y, s * v.z);
}

inline Vector3 operator*(const Vector3& v, double s) {
  return Vector3(s * v.x, s * v.y, s * v.z);
}

inline double operator*(const Vector3& a, const Vector3& b) {
  return a.x * b.x + a.y * b.y + a.z * b.z;
}

inline Vector3 operator^(const Vector3& a, const Vector3& b) {
  return Vector3(a.y * b.z - a.z * b.y,
		 a.z * b.x - a.x * b.z,
		 a.x * b.y - a.y * b.x);
}

inline bool operator==(const Vector3& a, const Vector3& b) {
  return a.x == b.x && a.y == b.y && a.z == b.z;
}

inline bool operator!=(const Vector3& a, const Vector3& b) {
  return a.x != b.x || a.y != b.y || a.z != b.z;
}

inline Vector3 operator/(const Vector3& v, double s) {
  double is = 1 / s;
  return Vector3(is * v.x, is * v.y, is * v.z);
}

inline ostream& operator<<(ostream& os, const Vector3& v) {
  os << "(" << v.x << ", " << v.y << ", " << v.z << ")";
  return os;
}

//
// **** Point3 operators ****

inline Vector3 operator-(const Point3& a, const Point3& b) {
  return Vector3(a.x - b.x, a.y - b.y, a.z - b.z);
}

inline bool operator==(const Point3& a, const Point3& b) {
  return a.x == b.x && a.y == b.y && a.z == b.z;
}

inline bool operator!=(const Point3& a, const Point3& b) {
  return a.x != b.x || a.y != b.y || a.z != b.z;
}

inline Point3 operator+(const Point3& p, const Vector3& v) {
  return Point3(p.x + v.x, p.y + v.y, p.z + v.z);
}

inline Point3 operator-(const Point3& p, const Vector3& v) {
  return Point3(p.x - v.x, p.y - v.y, p.z - v.z);
}

inline Point3 operator*(const Point3& p, double s) {
  return Point3(p.x * s, p.y * s, p.z * s);
}

inline Point3 operator*(double s, const Point3& p) {
  return Point3(p.x * s, p.y * s, p.z * s);
}

inline ostream& operator<<(ostream& os, const Point3& p) {
  os << "(" << p.x << ", " << p.y << ", " << p.z << ")";
  return os;
}

//
// **** Matrix3 operators ****

inline Vector3 operator*(const Matrix3& m, const Vector3& v) {
  return Vector3(m.r[0].x * v.x + m.r[0].y * v.y + m.r[0].z * v.z,
		 m.r[1].x * v.x + m.r[1].y * v.y + m.r[1].z * v.z,
		 m.r[2].x * v.x + m.r[2].y * v.y + m.r[2].z * v.z);
}

inline Vector3 operator*(const Vector3& v, const Matrix3& m) {
  return Vector3(m.r[0].x * v.x + m.r[1].x * v.y + m.r[2].x * v.z,
		 m.r[0].y * v.x + m.r[1].y * v.y + m.r[2].y * v.z,
		 m.r[0].z * v.x + m.r[1].z * v.y + m.r[2].z * v.z);
}

inline Point3 operator*(const Matrix3& m, const Point3& p) {
  return Point3(m.r[0].x * p.x + m.r[0].y * p.y + m.r[0].z * p.z,
		m.r[1].x * p.x + m.r[1].y * p.y + m.r[1].z * p.z,
		m.r[2].x * p.x + m.r[2].y * p.y + m.r[2].z * p.z);
}

inline Point3 operator*(const Point3& p, const Matrix3& m) {
  return Point3(m.r[0].x * p.x + m.r[1].x * p.y + m.r[2].x * p.z,
		m.r[0].y * p.x + m.r[1].y * p.y + m.r[2].y * p.z,
		m.r[0].z * p.x + m.r[1].z * p.y + m.r[2].z * p.z);
}

inline Matrix3 operator*(const Matrix3& a, const Matrix3& b) {
#define MATMUL(R, C) (a.r[R].x*b[0].C + a.r[R].y*b[1].C + a.r[R].z*b[2].C)
  return Matrix3(Vector3(MATMUL(0, x), MATMUL(0, y), MATMUL(0, z)),
		 Vector3(MATMUL(1, x), MATMUL(1, y), MATMUL(1, z)),
		 Vector3(MATMUL(2, x), MATMUL(2, y), MATMUL(2, z)));
#undef MATMUL
}

inline Matrix3 operator+(const Matrix3& a, const Matrix3& b) {
  return Matrix3(a.r[0] + b.r[0],
		 a.r[1] + b.r[1],
		 a.r[2] + b.r[2]);
}

inline ostream& operator<<(ostream& os, const Matrix3& m) {
  os << m.r[0] << endl;
  os << m.r[1] << endl;
  os << m.r[2] << endl;
  return os;
}

//
// **** Vector4 operators ****

inline bool operator==(const Vector4& a, const Vector4& b) {
  return a.x == b.x && a.y == b.y && a.z == b.z && a.w == b.w;
}

inline bool operator!=(const Vector4& a, const Vector4& b) {
  return a.x != b.x || a.y != b.y || a.z != b.z || a.w != b.w;
}

inline Vector4 operator+(const Vector4& a, const Vector4& b) {
  return Vector4(a.x + b.x, a.y + b.y, a.z + b.z, a.w + b.w);
}

inline Vector4 operator-(const Vector4& a, const Vector4& b) {
  return Vector4(a.x - b.x, a.y - b.y, a.z - b.z, a.w - b.w);
}

inline Vector4 operator*(double s, const Vector4& v) {
  return Vector4(s * v.x, s * v.y, s * v.z, s * v.w);
}

inline Vector4 operator*(const Vector4& v, double s) {
  return Vector4(s * v.x, s * v.y, s * v.z, s * v.w);
}

inline double operator*(const Vector4& a, const Vector4& b) {
  return a.x * b.x + a.y * b.y + a.z * b.z + a.w * b.w;
}

inline ostream& operator<<(ostream& os, const Vector4& v) {
  os << "(" << v.x << ", " << v.y << ", " << v.z << "," << v.w << ")";
  return os;
}

//
// **** Matrix4 operators ****

inline Vector3 operator*(const Matrix4& m, const Vector3& v) {
  return Vector3(m.r[0].x * v.x + m.r[0].y * v.y + m.r[0].z * v.z,
		 m.r[1].x * v.x + m.r[1].y * v.y + m.r[1].z * v.z,
		 m.r[2].x * v.x + m.r[2].y * v.y + m.r[2].z * v.z);
}

inline Vector3 operator*(const Vector3& v, const Matrix4& m) {
  return Vector3(m.r[0].x * v.x + m.r[1].x * v.y + m.r[2].x * v.z,
		 m.r[0].y * v.x + m.r[1].y * v.y + m.r[2].y * v.z,
		 m.r[0].z * v.x + m.r[1].z * v.y + m.r[2].z * v.z);
}

inline Point3 operator*(const Matrix4& m, const Point3& p) {
  return Point3(m.r[0].x * p.x + m.r[0].y * p.y + m.r[0].z * p.z + m.r[0].w,
		m.r[1].x * p.x + m.r[1].y * p.y + m.r[1].z * p.z + m.r[1].w,
		m.r[2].x * p.x + m.r[2].y * p.y + m.r[2].z * p.z + m.r[2].w);
}

inline Point3 operator*(const Point3& p, const Matrix4& m) {
  return Point3(m.r[0].x * p.x + m.r[1].x * p.y + m.r[2].x * p.z + m.r[3].x,
		m.r[0].y * p.x + m.r[1].y * p.y + m.r[2].y * p.z + m.r[3].y,
		m.r[0].z * p.x + m.r[1].z * p.y + m.r[2].z * p.z + m.r[3].z);
}

inline Vector4 operator*(const Matrix4& m, const Vector4& v) {
  return Vector4(m.r[0].x*v.x + m.r[0].y*v.y + m.r[0].z*v.z + m.r[0].w*v.w,
		 m.r[1].x*v.x + m.r[1].y*v.y + m.r[1].z*v.z + m.r[1].w*v.w,
		 m.r[2].x*v.x + m.r[2].y*v.y + m.r[2].z*v.z + m.r[2].w*v.w,
		 m.r[3].x*v.x + m.r[3].y*v.y + m.r[3].z*v.z + m.r[3].w*v.w);
}

inline Vector4 operator*(const Vector4& v, const Matrix4& m) {
  return Vector4(m.r[0].x*v.x + m.r[1].x*v.y + m.r[2].x*v.z + m.r[3].x*v.w,
		 m.r[0].y*v.x + m.r[1].y*v.y + m.r[2].y*v.z + m.r[3].y*v.w,
		 m.r[0].z*v.x + m.r[1].z*v.y + m.r[2].z*v.z + m.r[3].z*v.w,
		 m.r[0].w*v.x + m.r[1].w*v.y + m.r[2].w*v.z + m.r[3].w*v.w);
}

inline Matrix4 operator*(const Matrix4& a, const Matrix4& b) {
#define MATMUL(R, C) (a[R].x*b[0].C+a[R].y*b[1].C+a[R].z*b[2].C+a[R].w*b[3].C)
  return Matrix4(Vector4(MATMUL(0, x),MATMUL(0, y),MATMUL(0, z),MATMUL(0, w)),
		 Vector4(MATMUL(1, x),MATMUL(1, y),MATMUL(1, z),MATMUL(1, w)),
		 Vector4(MATMUL(2, x),MATMUL(2, y),MATMUL(2, z),MATMUL(2, w)),
		 Vector4(MATMUL(3, x),MATMUL(3, y),MATMUL(3, z),MATMUL(3, w)));
#undef MATMUL
}

inline Matrix4 operator+(const Matrix4& a, const Matrix4& b) {
  return Matrix4(a[0] + b[0], a[1] + b[1], a[2] + b[2], a[3] + b[3]);
}

inline ostream& operator<<(ostream& os, const Matrix4& m) {
  os << m.r[0] << endl;
  os << m.r[1] << endl;
  os << m.r[2] << endl;
  os << m.r[3] << endl;
  return os;
}

inline Matrix4 Matrix4::inverse() const {
  // Compute inverse using Gauss-Jordan elimination; caller is responsible
  // for ensuring that the matrix isn't singular.
  Matrix4 a(*this);
  Matrix4 b(Matrix4::identity());
  int i, j;
  int p;
  
  for (j = 0; j < 4; j++) {
    p = j;
    for (i = j + 1; i < 4; i++) {
      if (fabs(a.r[i][j]) > fabs(a.r[p][j]))
	p = i;
    }
    // Swap rows p and j
    Vector4 t = a.r[p];
    a.r[p] = a.r[j];
    a.r[j] = t;
    t = b.r[p];
    b.r[p] = b.r[j];
    b.r[j] = t;
    double s = a.r[j][j];  // if s == 0, the matrix is singular
    a.r[j] *= (1.0f / s);
    b.r[j] *= (1.0f / s);
    // Eliminate off-diagonal elements
    for (i = 0; i < 4; i++) {
      if (i != j) {
	b.r[i] -= a.r[i][j] * b.r[j];
	a.r[i] -= a.r[i][j] * a.r[j];
      }
    }
  }
  return b;
}

#endif /* _VECMATH_H_ */
